#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] can not import {__name__}.{__file__}')
#-----------------------------------------------------------------------------------------

#-----------------------------------------------------------------------------------------
# Basic imports
#-----------------------------------------------------------------------------------------
import os, sys, re, argparse, getpass, random
import re, importlib
from math import inf
import datetime
fnow = datetime.datetime.strftime
dnow = datetime.datetime.now
now = lambda: fnow(dnow(),"%Y-%m-%d %H:%M:%S")
# ------------------------------------------------------------------------------------------
# imports ----------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
try:
    from flask import Flask, render_template, request, redirect, url_for, session, abort, send_file
    from flask_wtf import FlaskForm
    from wtforms import SubmitField, MultipleFileField
    from werkzeug.utils import secure_filename
    from wtforms.validators import InputRequired
except: exit(f'[!] The required Flask packages missing:\tFlask>=3.0.2, Flask-WTF>=1.2.1\n  ⇒ pip install Flask==3.0.2 Flask-WTF==1.2.1')

try: from waitress import serve
except: exit(f'[!] The required waitress package missing:\twaitress>=3.0.0\n  ⇒ pip install waitress==3.0.0')
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------
g_break="+ - + - + - + - + - + - + - + - + - + - + - + - + - + - +"
print(f'{g_break}\nStarting...\n{g_break}')
#-----------------------------------------------------------------------------------------
# ==> Need to define working directories
#-----------------------------------------------------------------------------------------
PYDIR = os.path.dirname(__file__) # script directory of __main__.py

# this is for creating temporary html+css templates
# try parsing args
parser = argparse.ArgumentParser()
parser.add_argument('--dir', type=str, default='', help="path of workspace directory")
parser.add_argument('--cos', type=int, default=0, help="use 1 to create-on-start - create (overwrites) pages")
parser.add_argument('--coe', type=int, default=0, help="use 1 to clean-on-exit - deletes pages")
parsed = parser.parse_args()

# define working dir - contains all bases
WORKDIR = f'{parsed.dir}'
# if still not specified, set as default
if not WORKDIR: WORKDIR = os.path.join(os.getcwd()) 
print(f'↪ Workspace directory is {WORKDIR}')
try: os.makedirs(WORKDIR, exist_ok=True)
except: exit(f'[!] Workspace directory was not found and could not be created')
sys.path.append(WORKDIR)  # otherwise append to sys.path











CONFIG = 'current' # the config-dict to read from
CONFIG_MODULE = 'configs'  # the name of configs module
CONFIGS_FILE = f'{CONFIG_MODULE}.py' # the name of configs file

CSV_DELIM=','
SSV_DELIM='\n'
MAX_STR_LEN=50
CSV_DTYPE=f'U{MAX_STR_LEN*2}'

LOGIN_ORD = ['ADMIN','UID','NAME','PASS']

DEFAULT_USER = 'admin'
DEFAULT_ACCESS = f'DABU+'

#-----------------------------------------------------------------------------------------

#-----------------------------------------------------------------------------------------
# password policy
#-----------------------------------------------------------------------------------------
def VALIDATE_PASS(instr):   # a function that can validate the password - returns bool type
    try: assert (len(instr) < MAX_STR_LEN) and bool(re.fullmatch("(\w|@|\.)+", instr)) # alpha_numeric @.
    except AssertionError: return False
    return True
#-----------------------------------------------------------------------------------------
# uid policy
def VALIDATE_UID(instr):   # a function that can validate the uid - returns bool type
    try: assert (len(instr) < MAX_STR_LEN) and bool(re.fullmatch("(\w)+", instr)) # alpha_numeric 
    except AssertionError: return False
    return True
#-----------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------
# name policy
def VALIDATE_NAME(instr): return  (len(instr) >0) and (len(instr) < MAX_STR_LEN) and bool(re.fullmatch("((\w)(\w|\s)*(\w))|(\w)", instr)) # alpha-neumeric but no illegal spaces before or after
#-----------------------------------------------------------------------------------------

# this is useful for docker
# we try to read args from os env variables
def DEFAULT_CONFIG_GENERATE(env): return '\ndefault = {\n' + f"""
    'base' 		  :	"{env.get('base', '__base__')}",
    'secret'	  :	"{env.get('secret', '__secret__.txt')}",
    'login'       : "{env.get('login', '__login__.csv')}",
    'rename'      : {int(env.get('rename', 0))},
    'topic'       : "{env.get('topic', 'tOpIcS')}",
    'emoji'       : "{env.get('emoji', '◉')}",
    'welcome'     : "{env.get('welcome', 'Welcome!')}",
    'case'        : {int(env.get('case', 0))},
    'ext'         : "{env.get('ext', '')}",
    'required'    : "{env.get('required', '')}",
    'maxupcount'  : {int(env.get('maxupcount', -1))},
    'maxupsize'   : "{env.get('maxupsize', '40GB')}",
    'maxconnect'  : {int(env.get('maxconnect', 150))},
    'port'        : "{env.get('port', '8080')}",
    'host'        : "{env.get('host', '0.0.0.0')}",
    'uploads'     : "{env.get('uploads', '__uploads__')}",
    'downloads'   : "{env.get('downloads', '__downloads__')}",
    'archives'    : "{env.get('archives', '__archives__')}",
    'board'       : "{env.get('board', '__board__.ipynb')}",
    'threads'     : {int(env.get('threads', 4))},
    'verbose'     : {int(env.get('verbose', 2))},
""" + '\n}\n\ncurrent = default\n\n'


def DEFAULT_CONFIG_WRITE(file_path, env):
    with open(file_path, 'w', encoding='utf-8') as f: f.write(DEFAULT_CONFIG_GENERATE(env))



def DICT2CSV(path, d):
    with open(path, 'w', encoding='utf-8') as f: 
        f.write(CSV_DELIM.join(LOGIN_ORD)+SSV_DELIM)
        for v in d.values(): f.write(CSV_DELIM.join(v)+SSV_DELIM)

def CSV2DICT(path):
    with open(path, 'r', encoding='utf-8') as f: 
        s = f.read()
        lines = s.split(SSV_DELIM)
        d = dict()
        for line in lines[1:]:
            if line:
                cells = line.split(CSV_DELIM)
                d[f'{cells[1]}'] = cells
        return d

def GET_SECRET_KEY(postfix):
    return '{}:{}'.format(random.randint(1111111111, 9999999999), postfix)


def GET_VALID_RE_PATTERN(validext):
    if not validext: return ".+"
    pattern=""
    for e in validext: pattern+=f'{e}|'
    return pattern[:-1]


def CREATE_LOGIN_FILE(login_xl_path):  
    this_user = getpass.getuser()
    if not (VALIDATE_UID(this_user)):  this_user=DEFAULT_USER
    DICT2CSV(login_xl_path, { f'{this_user}' : [DEFAULT_ACCESS,  f'{this_user}', f'{this_user}', f''] } ) # save updated login information to csv
    return this_user



def READ_DB_FROM_DISK(login_xl_path):
    try: return CSV2DICT(login_xl_path), True
    except: return dict(), False
# ------------------------------------------------------------------------------------------
def WRITE_DB_TO_DISK(login_xl_path, db_frame): # will change the order
    try:
        DICT2CSV(login_xl_path, db_frame) # save updated login information to csv
        return True
    except PermissionError:
        return False
    


def GET_FILE_LIST (d): 
    dlist = []
    #d = DOWNLOAD_FOLDER_PATH, ARCHIVE_FOLDER_PATH
    for f in os.listdir(d):
        p = os.path.join(d, f)
        if os.path.isfile(p): dlist.append(f)
    return dlist


def DISPLAY_SIZE_READABLE(mus):
    # find max upload size in appropiate units
    mus_kb = mus/(2**10)
    if len(f'{int(mus_kb)}') < 4:
        mus_display = f'{mus_kb:.2f} KB'
    else:
        mus_mb = mus/(2**20)
        if len(f'{int(mus_mb)}') < 4:
            mus_display = f'{mus_mb:.2f} MB'
        else:
            mus_gb = mus/(2**30)
            if len(f'{int(mus_gb)}') < 4:
                mus_display = f'{mus_gb:.2f} GB'
            else:
                mus_tb = mus/(2**40)
                mus_display = f'{mus_tb:.2f} TB'
    return mus_display


def NEW_NOTEBOOK_STR(title, nbformat=4, nbformat_minor=2):
    return '{"cells": [{"cell_type": "markdown","metadata": {},"source": [ "'+str(title)+'" ] } ], "metadata": { }, "nbformat": '+str(nbformat)+', "nbformat_minor": '+str(nbformat_minor)+'}'


#-----------------------------------------------------------------------------------------
# Special Objects
#-----------------------------------------------------------------------------------------
class Fake:
    def __len__(self): return len(self.__dict__)
    def __init__(self, **kwargs) -> None:
        for name, attribute in kwargs.items():  setattr(self, name, attribute)
#-----------------------------------------------------------------------------------------

# ******************************************************************************************


































#-----------------------------------------------------------------------------------------
# try to import configs
# inside the WORKDIR there should be 'configs.py' file
# check if 'configs.py` exsists or not`
CONFIGS_FILE_PATH = os.path.join(WORKDIR, CONFIGS_FILE) # should exsist under workdir
if not os.path.isfile(CONFIGS_FILE_PATH):
    print(f'↪ Creating default config "{CONFIGS_FILE}" from environment...')
    DEFAULT_CONFIG_WRITE(CONFIGS_FILE_PATH, dict(os.environ))
try: 
    c_module = importlib.import_module(f'{CONFIG_MODULE}')
    print(f'↪ Imported config-module "{CONFIG_MODULE}" from {c_module.__file__}')
except: exit(f'[!] Could import configs module "{CONFIG_MODULE}" at "{CONFIGS_FILE_PATH[:-3]}"')
try:
    print(f'↪ Reading config from {CONFIG_MODULE}.{CONFIG}')
    config_dict = getattr(c_module, CONFIG)
    print(f'  ↦ type:{type(config_dict)}')
except:
    exit(f'[!] Could not read config from {CONFIG_MODULE}.{CONFIG}')

if not isinstance(config_dict, dict): 
    try: config_dict=config_dict()
    except: pass
if not isinstance(config_dict, dict): raise exit(f'Expecting a dict object for config')

try: 
    print(f'↪ Building config from {CONFIG_MODULE}.{CONFIG}')
    #for k,v in config_dict.items(): print(f'  ↦ {k}\t\t{v}')
    args = Fake(**config_dict)
except: exit(f'[!] Could not read config')
if not len(args): exit(f'[!] Empty or Invalid config provided')
# for k in default_args: 
#     if not hasattr(args, k): exit(f'[!] config is missing attribute ({k})')
# ******************************************************************************************

# ------------------------------------------------------------------------------------------
# verbose levels
# dprints for user level logs
# sprint  for sever level logs
# ------------------------------------------------------------------------------------------

if args.verbose==0: # no log
    def sprint(msg): pass
    def dprint(msg): pass
elif args.verbose==1: # only server logs
    def sprint(msg): print(msg)
    def dprint(msg): pass 
elif args.verbose==2: # server and user logs
    def sprint(msg): print(msg)
    def dprint(msg): print(f'[{now()}]\t{msg}' )
else:
    def sprint(msg): pass
    def dprint(msg): pass
# ------------------------------------------------------------------------------------------














































# ------------------------------------------------------------------------------------------
# html pages
# ------------------------------------------------------------------------------------------
#-----------------------------------------------------------------------------------------
# Create HTML
# ------------------------------------------------------------------------------------------
# ******************************************************************************************
HTML_TEMPLATES = dict(
# ******************************************************************************************
board="""""",
# ******************************************************************************************
admin = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> ⚙ {{ config.topic }} | {{ session.uid }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">					 
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    
    <div align="left" style="padding: 20px;">
        <div class="topic_mid">{{ config.topic }}</div>
        <div class="userword">{{session.uid}} {{ session.emojid }} {{session.named}}</div>
        <br>
        <a href="{{ url_for('logout') }}" class="btn_logout">Logout</a>
        <a href="{{ url_for('upload') }}" class="btn_back">Back</a>
        <a href="{{ url_for('adminpage') }}" class="btn_refresh">Refresh</a>
        <br>
        <br>

        {% if success %}
        <span class="admin_mid" style="animation-name: fader_admin_success;">✓ {{ status }} </span>
        {% else %}
        <span class="admin_mid" style="animation-name: fader_admin_failed;">✗ {{ status }} </span>
        {% endif %}
        <br>
        <br>
        {% if '+' in session.admind %}
        <a href="{{ url_for('adminpage',req_cmd='dbw') }}" class="btn_admin">⚙ Persist login-db ↷</a>
        <br>
        <br>
        <a href="{{ url_for('adminpage',req_cmd='dbr') }}" class="btn_admin">⚙ Reload login-db ↺</a>
        <br>
        <br>
        <a href="{{ url_for('adminpage',req_cmd='upd') }}" class="btn_admin">⚙ Update download-list ▤</a>
        <br>
        <br>
        <a href="{{ url_for('adminpage',req_cmd='upa') }}" class="btn_admin">⚙ Update archive-list ⊞</a>
        <br>
        <br>
        <a href="{{ url_for('adminpage',req_cmd='ref') }}" class="btn_admin">⚙ Refresh board ▣</a>
        <br>
        <br>
        <button class="btn_admin" onclick="confirm_repass()">⚙ Reset Password ⨝</button>
            <script>
                function confirm_repass() {
                let res = prompt("Enter UID", ""); 
                if (res != null) {
                    location.href = "{{ url_for('repass',req_uid='::::') }}".replace("::::", res);
                    }
                }
            </script>
        <br>
        <br>
        {% endif %}
    </div>
            
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
# ******************************************************************************************
login = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> 🔒 {{ config.topic }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">  
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->

    <div align="center">
        <br>
        <div class="topic">{{ config.topic }}</div>
        <br>
        <br>
        <form action="{{ url_for('login') }}" method="post">
            <br>
            <div style="font-size: x-large;">{{ warn }}</div>
            <br>
            <div class="msg_login">{{ msg }}</div>
            <br>
            <input id="uid" name="uid" type="text" placeholder="... user-id ..." class="txt_login"/>
            <br>
            <br>
            <input id="passwd" name="passwd" type="password" placeholder="... password ..." class="txt_login"/>
            <br>
            <br>
            {% if config.rename>0 %}
            <input id="named" name="named" type="text" placeholder="... update-name ..." class="txt_login"/>
            {% if config.rename>1 %}
            <input id="emojid" name="emojid" type="text" placeholder={{ config.emoji }} class="txt_login_small"/>
            {% endif %}
            <br>
            {% endif %}
            <br>
            <input type="submit" class="btn_login" value="Login"> 
            <br>
            <br>
        </form>
    </div>

    <!-- ---------------------------------------------------------->
    
    <div align="center">
    <div>
    <span style="font-size: xx-large;">{{ config.emoji }}</span>
    <br>
    
    </div>
    <!-- <a href="https://emojipicker.com/" target="_blank" class="btn_login">...</a> -->
    <!--<div style="font-size:large"><a href="https://github.com/NelsonSharma/topics"  target="_blank"> 📤 📥 </a></div>-->
    <br>
    </div>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
# ******************************************************************************************
download = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> 🔻 {{ config.topic }} | {{ session.uid }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">           
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    
    <div align="left" style="padding: 20px;">
        <div class="topic_mid">{{ config.topic }}</div>
        <div class="userword">{{session.uid}} {{ session.emojid }} {{session.named}}</div>
        <br>
        <a href="{{ url_for('logout') }}" class="btn_logout">Logout</a>
        <a href="{{ url_for('upload') }}" class="btn_back">Back</a>
        <a href="{{ url_for('download') }}" class="btn_refresh">Refresh</a>
        <br>
        <br>
        <div class="files_status">Downloads</div>
        <br>
        <div class="files_list_down">
            <ol>
            {% for file in config.dfl %}
            <li><a href="{{ (request.path + '/' if request.path != '/' else '') + file }}" style="text-decoration: none; color: rgb(20, 20, 20);" >{{ file }}</a></li>
            <br>
            {% endfor %}
            </ol>
        </div>
        <br>
        <br>
    </div>

    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
# ******************************************************************************************
archive = """
<html>
    <head>
        <meta charset="UTF-8">
        <title> 🔶 {{ config.topic }} | {{ session.uid }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">           
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    
    <div align="left" style="padding: 20px;">
        <div class="topic_mid">{{ config.topic }}</div>
        <div class="userword">{{session.uid}} {{ session.emojid }} {{session.named}}</div>
        <br>
        <a href="{{ url_for('logout') }}" class="btn_logout">Logout</a>
        <a href="{{ url_for('upload') }}" class="btn_back">Back</a>
        <a href="{{ url_for('archive') }}" class="btn_refresh">Refresh</a>
        <br>
        <br>
        <div class="files_status">Archives</div>
        <br>
        <div class="files_list_down">
            <ol>
            {% for file in config.afl %}
            <li><a href="{{ (request.path + '/' if request.path != '/' else '') + file }}" style="text-decoration: none; color: rgb(20, 20, 20);" >{{ file }}</a></li>
            <br>
            {% endfor %}
            </ol>
        </div>
        <br>
        <br>
    </div>

    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
# ******************************************************************************************
upload="""
<html>
    <head>
        <meta charset="UTF-8">
        <title> 🔺 {{ config.topic }} | {{ session.uid }} </title>
        <link rel="stylesheet" href="{{ url_for('static', filename='style.css') }}">					 
    </head>
    <body>
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    
    <div align="left" style="padding: 20px;">
        <div class="topic_mid">{{ config.topic }}</div>
        <div class="userword">{{session.uid}} {{ session.emojid }} {{session.named}}</div>
        <br>
        <a href="{{ url_for('logout') }}" class="btn_logout">Logout</a>
        {% if "D" in session.admind %}
        <a href="{{ url_for('download') }}" class="btn_download">Downloads</a>
        {% endif %}
        {% if "A" in session.admind %}
        <a href="{{ url_for('archive') }}" class="btn_archive">Archives</a>
        {% endif %}
        {% if "U" in session.admind %}
        <a href="{{ url_for('uploadf') }}" class="btn_refresh">Refresh</a>
        {% endif %}
        {% if "B" in session.admind and config.board %}
        <a href="{{ url_for('board') }}" class="btn_board" target="_blank">Board</a>
        {% endif %}
        {% if '+' in session.admind %}
        <a href="{{ url_for('adminpage') }}" class="btn_admin">⚙</a>
        {% endif %}
        <br>
        <br>
        {% if config.muc!=0 and "U" in session.admind %}
        <div class="status">
            <ol>
            {% for s,f in status %}
            {% if s %}
            {% if s<0 %}
            <li style="color: #ffffff;">{{ f }}</li>
            {% else %}
            <li style="color: #47ff6f;">{{ f }}</li>
            {% endif %}
            {% else %}
            <li style="color: #ff6565;">{{ f }}</li>
            {% endif %}
            {% endfor %}
            </ol>
        </div>
        <br>
        <form method='POST' enctype='multipart/form-data'>
            {{form.hidden_tag()}}
            {{form.file()}}
            {{form.submit()}}
        </form>
        <br>
        <div class="files_status">Uploads
        <button class="btn_purge" onclick="confirm_purge()">Purge</button>
            <script>
                function confirm_purge() {
                let res = confirm("Purge all the uploaded files now?");
                if (res == true) {
                    location.href = "{{ url_for('purge') }}";
                    }
                }
            </script>
        </div>
        <br>
        <div class="files_list_up">
            <ol>
            {% for f in session.filed %}
            <li>{{ f }}</li>
            {% endfor %}
            </ol>
        </div>
        <br>
        <br>
        {% endif %}
    </div>
            
    <!-- ---------------------------------------------------------->
    </br>
    <!-- ---------------------------------------------------------->
    </body>
</html>
""",
# ******************************************************************************************
)
CSS_TEMPLATES = dict(
# ******************************************************************************************
style = """
.github_info {
    padding: 2px 10px;
    background-color: #516fa7; 
    color: #ffffff;
    font-size: medium;
    font-weight: bold;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.topic{
    color: #000000;
    font-size: xxx-large;
    font-weight: bold;
    font-family:monospace;    
}


.msg_login{
    color: #060472; 
    font-size: large;
    font-weight: bold;
    font-family:monospace;    
    animation-duration: 3s; 
    animation-name: fader_msg;
}
@keyframes fader_msg {from {color: #ffffff;} to {color: #060472; } }


.txt_login{

    text-align: center;
    font-family:monospace;

    box-shadow: inset #abacaf 0 0 0 2px;
    border: 0;
    background: rgba(0, 0, 0, 0);
    appearance: none;
    position: relative;
    border-radius: 3px;
    padding: 9px 12px;
    line-height: 1.4;
    color: rgb(0, 0, 0);
    font-size: 16px;
    font-weight: 400;
    height: 40px;
    transition: all .2s ease;
    :hover{
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 2px;
    }
    :focus{
        background: #fff;
        outline: 0;
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 3px;
    }
}
::placeholder {
    color: #888686;
    opacity: 1;
    font-weight: bold;
    font-style: oblique;
    font-family:monospace;   
}


.txt_login_small{
    box-shadow: inset #abacaf 0 0 0 2px;
    text-align: center;
    font-family:monospace;
    border: 0;
    background: rgba(0, 0, 0, 0);
    appearance: none;
    position: absolute;
    border-radius: 3px;
    padding: 9px 12px;
    margin: 0px 0px 0px 4px;
    line-height: 1.4;
    color: rgb(0, 0, 0);
    font-size: 16px;
    font-weight: 400;
    height: 40px;
    width: 45px;
    transition: all .2s ease;
    :hover{
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 2px;
    }
    :focus{
        background: #fff;
        outline: 0;
        box-shadow: 0 0 0 0 #fff inset, #1de9b6 0 0 0 3px;
    }
}

.topic_mid{
    color: #000000;
    font-size: x-large;
    font-style: italic;
    font-weight: bold;
    font-family:monospace;    
}

.userword{
    color: #12103c;
    font-weight: bold;
    font-family:monospace;    
    font-size: xxx-large;
}

.status{
    padding: 10px 10px;
    background-color: #232323; 
    color: #ffffff;
    font-size: medium;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.files_status{
    font-weight: bold;
    font-size: x-large;
    font-family:monospace;
}

.files_list_up{
    padding: 10px 10px;
    background-color: #ececec; 
    color: #080000;
    font-size: medium;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.files_list_down{
    padding: 10px 10px;
    background-color: #ececec; 
    color: #080000;
    font-size: x-large;
    font-weight: bold;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.btn_logout {
    padding: 2px 10px 2px;
    background-color: #060472; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.btn_refresh {
    padding: 2px 10px 2px;
    background-color: #a19636; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.btn_purge {
    padding: 2px 10px 2px;
    background-color: #9a0808; 
    border-style: none;
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.btn_admin {
    padding: 2px 10px 2px;
    background-color: #000000; 
    border-style: none;
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.btn_board {
    padding: 2px 10px 2px;
    background-color: #934377; 
    border-style: none;
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.btn_login {
    padding: 2px 10px 2px;
    background-color: #060472; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
    border-style:  none;
}

.btn_download {
    padding: 2px 10px 2px;
    background-color: #089a28; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.btn_archive{
    padding: 2px 10px 2px;
    background-color: #10a58a; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}

.btn_back {
    padding: 2px 10px 2px;
    background-color: #48952d; 
    color: #FFFFFF;
    font-weight: bold;
    font-size: large;
    border-radius: 10px;
    font-family:monospace;
    text-decoration: none;
}


.admin_mid{
    color: #000000; 
    font-size: x-large;
    font-weight: bold;
    font-family:monospace;    
    animation-duration: 10s;
}
@keyframes fader_admin_failed {from {color: #ff0000;} to {color: #000000; } }
@keyframes fader_admin_success {from {color: #22ff00;} to {color: #000000; } }
"""
)
# ******************************************************************************************

#try:
    
TEMPLATES_DIR, STATIC_DIR = os.path.join(PYDIR, "templates"), os.path.join(PYDIR, "static")

os.makedirs(TEMPLATES_DIR, exist_ok=True)
print(f'↪ Creating html templates @ {TEMPLATES_DIR}')
for k,v in HTML_TEMPLATES.items():
    h = os.path.join(TEMPLATES_DIR, f"{k}.html")
    if (not os.path.isfile(h)) or bool(parsed.cos):
        print(f'  ↦ page:{h}')
        with open(h, 'w', encoding='utf-8') as f: f.write(v)

os.makedirs(STATIC_DIR, exist_ok=True)
print(f'↪ Creating css templates @ {STATIC_DIR}')
for k,v in CSS_TEMPLATES.items():
    h = os.path.join(STATIC_DIR, f"{k}.css")
    if (not os.path.isfile(h)) or bool(parsed.cos):
        print(f'  ↦ page:{h}')
        with open(h, 'w', encoding='utf-8') as f: f.write(v)

print(f'↪ Created html/css templates @ {PYDIR}')
#except: exit(f'[!] could not create html templates at {PYDIR}')
# topics app 













































    
# Read base dir first 
BASEDIR = os.path.abspath(((os.path.join(WORKDIR, args.base)) if args.base else WORKDIR))
try:     os.makedirs(BASEDIR, exist_ok=True)
except:  exit(f'[!] base directory  @ {BASEDIR} was not found and could not be created') 
sprint(f'⚙ Base dicectiry:\t{BASEDIR}')
# ------------------------------------------------------------------------------------------
# WEB-SERVER INFORMATION
# ------------------------------------------------------------------------------------------\
if not args.secret: exit(f'[!] secret key was not provided!')
APP_SECRET_KEY_FILE = os.path.join(BASEDIR, args.secret)
if not os.path.isfile(APP_SECRET_KEY_FILE): #< --- if key dont exist, create it
    APP_SECRET_KEY =  GET_SECRET_KEY(now())
    try:
        with open(APP_SECRET_KEY_FILE, 'w') as f: f.write(APP_SECRET_KEY) #<---- auto-generated key
    except: exit(f'[!] could not create secret key @ {APP_SECRET_KEY_FILE}')
    sprint(f'⇒ New secret created:\t{APP_SECRET_KEY_FILE}')
else:
    try:
        with open(APP_SECRET_KEY_FILE, 'r') as f: APP_SECRET_KEY = f.read()
        sprint(f'⇒ Loaded secret file:\t{APP_SECRET_KEY_FILE}')
    except: exit(f'[!] could not read secret key @ {APP_SECRET_KEY_FILE}')


# ------------------------------------------------------------------------------------------
# LOGIN DATABASE - CSV
# ------------------------------------------------------------------------------------------
if not args.login: exit(f'[!] login file was not provided!')    
LOGIN_XL_PATH = os.path.join( BASEDIR, args.login) 
if not os.path.isfile(LOGIN_XL_PATH): 
    sprint(f'⇒ Creating new login file:\t{LOGIN_XL_PATH}')
    this_user = CREATE_LOGIN_FILE(LOGIN_XL_PATH)
    sprint(f'⇒ Created new login with user "{this_user}" at file:\t{LOGIN_XL_PATH}')

def read_db_from_disk():
    db_frame, res = READ_DB_FROM_DISK(LOGIN_XL_PATH)
    if res: sprint(f'⇒ Loaded login file:\t{LOGIN_XL_PATH}')
    else: sprint(f'⇒ Failed reading login file:\t{LOGIN_XL_PATH}')
    return db_frame
# ------------------------------------------------------------------------------------------
def write_db_to_disk(db_frame): # will change the order
    res = WRITE_DB_TO_DISK(LOGIN_XL_PATH, db_frame)
    if res: sprint(f'⇒ Persisted login file:\t{LOGIN_XL_PATH}')
    else:  sprint(f'⇒ PermissionError - {LOGIN_XL_PATH} might be open, close it first.')
    return res
# ------------------------------------------------------------------------------------------
db = read_db_from_disk()  #<----------- Created db here 
# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# download settings
# ------------------------------------------------------------------------------------------
if not args.downloads: exit(f'[!] downloads folder was not provided!')
DOWNLOAD_FOLDER_PATH = os.path.join( BASEDIR, args.downloads) 
try: os.makedirs(DOWNLOAD_FOLDER_PATH, exist_ok=True)
except: exit(f'[!] downloads folder @ {DOWNLOAD_FOLDER_PATH} was not found and could not be created')
sprint(f'⚙ Download Folder:\t{DOWNLOAD_FOLDER_PATH}') 
#sprint(f'⚙ Download filelist\t{len(DOWNLOAD_FILE_LIST)} item(s)')
# ------------------------------------------------------------------------------------------
# archive settings
# ------------------------------------------------------------------------------------------
if not args.archives: exit(f'[!] archives folder was not provided!')
ARCHIVE_FOLDER_PATH = os.path.join( BASEDIR, args.archives) 
try: os.makedirs(ARCHIVE_FOLDER_PATH, exist_ok=True)
except: exit(f'[!] archives folder @ {ARCHIVE_FOLDER_PATH} was not found and could not be created')
sprint(f'⚙ Archive Folder:\t{ARCHIVE_FOLDER_PATH}')
#sprint(f'⚙ Archive filelist\t{len(ARCHIVE_FILE_LIST)} item(s)')


# ------------------------------------------------------------------------------------------
# upload settings
# ------------------------------------------------------------------------------------------
if not args.uploads: exit(f'[!] uploads folder was not provided!')
UPLOAD_FOLDER_PATH = os.path.join( BASEDIR, args.uploads ) 
try: os.makedirs(UPLOAD_FOLDER_PATH, exist_ok=True)
except: exit(f'[!] uploads folder @ {UPLOAD_FOLDER_PATH} was not found and could not be created')
sprint(f'⚙ Upload Folder:\t{UPLOAD_FOLDER_PATH}')


# ------------------------------------------------------------------------------------------
# validation
# ------------------------------------------------------------------------------------------

ALLOWED_EXTENSIONS = set([x.strip() for x in args.ext.split(',') if x])  # a set or list of file extensions that are allowed to be uploaded 
if '' in ALLOWED_EXTENSIONS: ALLOWED_EXTENSIONS.remove('')
VALID_FILES_PATTERN = GET_VALID_RE_PATTERN(ALLOWED_EXTENSIONS)
REQUIRED_FILES = set([x.strip() for x in args.required.split(',') if x])  # a set or list of file extensions that are allowed to be uploaded 
if '' in REQUIRED_FILES: REQUIRED_FILES.remove('')
def VALIDATE_FILENAME(filename):   # a function that checks for valid file extensions based on ALLOWED_EXTENSIONS
    if '.' in filename: 
        name, ext = filename.rsplit('.', 1)
        safename = f'{name}.{ext.lower()}'
        if REQUIRED_FILES:  isvalid = (safename in REQUIRED_FILES)
        else:               isvalid = bool(re.fullmatch(f'.+\.({VALID_FILES_PATTERN})$', safename))
    else:               
        name, ext = filename, ''
        safename = f'{name}'
        if REQUIRED_FILES:  isvalid = (safename in REQUIRED_FILES)
        else:               isvalid = (not ALLOWED_EXTENSIONS)
    return isvalid, safename

def str2bytes(size):
    sizes = dict(KB=2**10, MB=2**20, GB=2**30, TB=2**40)
    return int(float(size[:-2])*sizes.get(size[-2:].upper(), 0))
MAX_UPLOAD_SIZE = str2bytes(args.maxupsize)     # maximum upload file size 
MAX_UPLOAD_COUNT = ( inf if args.maxupcount<0 else args.maxupcount )       # maximum number of files that can be uploaded by one user
INITIAL_UPLOAD_STATUS = []           # a list of notes to be displayed to the users about uploading files
if REQUIRED_FILES: INITIAL_UPLOAD_STATUS.append((-1, f'accepted files [{len(REQUIRED_FILES)}]:\t{REQUIRED_FILES}'))
else:
    if ALLOWED_EXTENSIONS:  INITIAL_UPLOAD_STATUS.append((-1, f'allowed extensions [{len(ALLOWED_EXTENSIONS)}]:\t{ALLOWED_EXTENSIONS}'))
INITIAL_UPLOAD_STATUS.append((-1, f'max upload size:\t{DISPLAY_SIZE_READABLE(MAX_UPLOAD_SIZE)}'))
if not (MAX_UPLOAD_COUNT is inf): INITIAL_UPLOAD_STATUS.append((-1, f'max upload count:\t{MAX_UPLOAD_COUNT}'))
sprint(f'⚙ Upload Settings ({len(INITIAL_UPLOAD_STATUS)})')
for s in INITIAL_UPLOAD_STATUS: sprint(f' ⇒ {s[1]}')
# ------------------------------------------------------------------------------------------
# download settings

# ------------------------------------------------------------------------------------------
BOARD_FILE_MD = None
BOARD_PAGE = ""
if args.board:
    try: 
        from nbconvert import HTMLExporter 
        has_nbconvert_package=True
    except:
        sprint(f'⚙ Board File is specified but requires nbconvert>=7.16.2 which is missing\n  ⇒ pip install nbconvert==7.16.2')
        has_nbconvert_package = False
    finally:
        if has_nbconvert_package:
            BOARD_FILE_MD = os.path.join(BASEDIR, f'{args.board}')
            if  os.path.isfile(BOARD_FILE_MD): sprint(f'⚙ Board File:\t{BOARD_FILE_MD}')
            else: 
                sprint(f'⚙ Board File:\t{BOARD_FILE_MD} not found - trying to create...')
                try:
                    with open(BOARD_FILE_MD, 'w', encoding='utf-8') as f: f.write(NEW_NOTEBOOK_STR(f'# Board 🔰 {args.topic}'))
                    sprint(f'⚙ Board File:\t{BOARD_FILE_MD} was created successfully!')
                except: 
                     BOARD_FILE_MD = None
                     sprint(f'⚙ Board File:\t{BOARD_FILE_MD} could not be created - Board will not be available!')
        else: BOARD_FILE_MD = None
if not BOARD_FILE_MD:  sprint(f'⚙ Board File:\tNot Enabled')
def update_board(): 
    global BOARD_PAGE
    res = False
    if BOARD_FILE_MD:
        try: 
            page,_ = HTMLExporter(template_name="classic").from_file(BOARD_FILE_MD, {'metadata':{'name':f'🔰 Board | {args.topic}'}}) 
            BOARD_PAGE = page
            sprint(f'⚙ Board File was updated:\t{BOARD_FILE_MD}')
            res=True
        except: 
            BOARD_PAGE=""
            sprint(f'⚙ Board File could not be updated:\t{BOARD_FILE_MD}')
    else: BOARD_PAGE=""
    return res

_ = update_board()


# ------------------------------------------------------------------------------------------
# application setting and instance
# ------------------------------------------------------------------------------------------
app = Flask(__name__)
app.secret_key =          APP_SECRET_KEY
app.config['base'] =      BASEDIR
app.config['uploads'] =   UPLOAD_FOLDER_PATH
app.config['downloads'] = DOWNLOAD_FOLDER_PATH
app.config['archives'] =  ARCHIVE_FOLDER_PATH
app.config['emoji'] =     args.emoji
app.config['topic'] =     args.topic
app.config['dfl'] =       GET_FILE_LIST(DOWNLOAD_FOLDER_PATH)
app.config['afl'] =       GET_FILE_LIST(ARCHIVE_FOLDER_PATH)
app.config['rename'] =    int(args.rename)
app.config['muc'] =       MAX_UPLOAD_COUNT
app.config['board'] =     (BOARD_FILE_MD is not None)

# ------------------------------------------------------------------------------------------
class UploadFileForm(FlaskForm): # The upload form using FlaskForm
    file = MultipleFileField("File", validators=[InputRequired()])
    submit = SubmitField("Upload File")
# ------------------------------------------------------------------------------------------

#%% [4]
# app.route  > all app.route implemented here 
# ------------------------------------------------------------------------------------------
# login
# ------------------------------------------------------------------------------------------
@app.route('/', methods =['GET', 'POST'])
def login():
    LOGIN_NEED_TEXT =       '🔒'
    LOGIN_FAIL_TEXT =       '❌'     
    LOGIN_NEW_TEXT =        '🔥'
    LOGIN_CREATE_TEXT =     '🔑'    
    #NAME, PASS = 2, 3
    global db#, HAS_PENDING#<--- only when writing to global wariables
    if request.method == 'POST' and 'uid' in request.form and 'passwd' in request.form:
        in_uid = f"{request.form['uid']}"
        in_passwd = f"{request.form['passwd']}"
        in_name = f'{request.form["named"]}' if 'named' in request.form else ''
        in_emoji = f'{request.form["emojid"]}' if 'emojid' in request.form else app.config['emoji']
        if ((not in_emoji) or (app.config['rename']<2)): in_emoji = app.config['emoji']
        in_query = in_uid if not args.case else (in_uid.upper() if args.case>0 else in_uid.lower())
        valid_query, valid_name = VALIDATE_UID(in_query) , VALIDATE_NAME(in_name)
        if not valid_query : record=None
        else: record = db.get(in_query, None) #print(f"◦ login attempt by [{in_uid}] will case-query [{in_query}]")
        #print(f"◦ record matched? [{record}]")
        if record is not None: 
            admind, uid, named, passwd = record
            admind = f'{admind}'.upper()
            #print(f"◦ matched record [{uid}|{named}] ")
            if not passwd: # fist login
                #print(f"◦ first login")
                if in_passwd: # new password provided
                    #print(f"[---------] new password provided [{in_passwd}]")
                    if VALIDATE_PASS(in_passwd): # new password is valid
                        db[uid][3]=in_passwd 
                        #HAS_PENDING+=1
                        if in_name!=named and valid_name and (app.config['rename']>0) : 
                            db[uid][2]=in_name
                            #HAS_PENDING+=1
                            dprint(f'⇒ {uid} ◦ {named} updated name to "{in_name}"') 
                            named = in_name
                        else:
                            if in_name: dprint(f'⇒ {uid} ◦ {named} provided invalid name "{in_name}" (will not update)') 
                        
                        #print(f'◦ updated record') # \n{record}

                        warn = LOGIN_CREATE_TEXT
                        msg = f'[{in_uid}] ({named}) New password was created successfully'
                        dprint(f'● {in_uid} {in_emoji} {named} just joined')
           
                    else: # new password is invalid valid
                        #print(f"[........] new password is invalid")  
                        warn = LOGIN_NEW_TEXT
                        msg=f'[{in_uid}] New password is invalid - can use alpha-numeric, underscore and @-symbol'
                        
                                               
                else: #new password not provided       
                    #print(f"[.....] new password was not provided")             
                    warn = LOGIN_NEW_TEXT
                    msg = f'[{in_uid}] New password required - can use alpha-numeric, underscore and @-symbol'
                                           
            else: # re login
                #print(f"◦ revist login")
                if in_passwd: # password provided
                    #print(f"[........] password provided {in_passwd}")  
                    if in_passwd==passwd:
                        #print(f"[......] password does match") 
                        folder_name = os.path.join(app.config['uploads'], uid) 
                        #print(f"user {session['uid']} landed on upload page")
                        try:
                            os.makedirs(folder_name, exist_ok=True)
                            #print(f"..... has directory {folder_name}")
                        except:
                            dprint(f'✗ directory could not be created @ {folder_name} :: Force logout user {uid}')
                            session['has_login'] = False
                            session['uid'] = uid
                            session['named'] = named
                            session['emojid'] = ''
                            return redirect(url_for('logout'))
                    
                        session['has_login'] = True
                        session['uid'] = uid
                        session['admind'] = admind
                        session['filed'] = os.listdir(folder_name)
                        session['emojid'] = in_emoji 
                        
                        if in_name!=named and  valid_name and  (app.config['rename']>0): 
                            session['named'] = in_name
                            db[uid][2] = in_name
                            #HAS_PENDING+=1
                            dprint(f'⇒ {uid} ◦ {named} updated name to "{in_name}"') 
                            named = in_name
                            #print(f'◦ updated record') # \n{record}
                        else: 
                            session['named'] = named
                            if in_name: dprint(f'⇒ {uid} ◦ {named} provided invalid name "{in_name}" (will not update)')  

                        #print(f'◦ login success {uid}|{named}')
                        dprint(f'● {session["uid"]} {session["emojid"]} {session["named"]} has logged in') 
                        #print(f"filed @ login= {session['filed']}")
                        return redirect(url_for('upload'))
                    else:  
                        #print(f"[.....] password does not match")  
                        warn = LOGIN_FAIL_TEXT
                        msg = f'[{in_uid}] Password mismatch'                  
                else: # password not provided
                    #print(f"[....] password not provided")  
                    warn = LOGIN_FAIL_TEXT
                    msg = f'[{in_uid}] Password not provided'
        else:
            #print(f"◦ unmatched record {in_uid}")
            warn = LOGIN_FAIL_TEXT
            msg = f'[{in_uid}] Not a valid user' 

    else:
        if session.get('has_login', False):  return redirect(url_for('upload'))
        #print(f"+ page hit")
        msg = args.welcome
        warn = LOGIN_NEED_TEXT 
        
    return render_template('login.html', msg = msg,  warn = warn)

@app.route('/logout')
def logout():
    r""" logout a user and redirect to login page """
    if not session.get('has_login', False):  return redirect(url_for('login'))
    if not session.get('uid', False): return redirect(url_for('login'))
    #print(f"◦ log out user {session['uid']}")
    if session['has_login']:  dprint(f'● {session["uid"]} {session["emojid"]} {session["named"]} has logged out') 
    else: dprint(f'✗ {session["uid"]} ◦ {session["named"]} was removed due to invalid uid ({session["uid"]})') 
    session['has_login'] = False
    session['uid'] = ""
    session['named'] = ""
    session['emojid'] = ""
    session['admind'] = ''
    session['filed'] = []
    return redirect(url_for('login'))
# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# board
# ------------------------------------------------------------------------------------------
@app.route('/board', methods =['GET'])
def board():
    if not session.get('has_login', False): return redirect(url_for('login'))
    if 'B' not in session['admind']:  return redirect(url_for('upload'))
    return BOARD_PAGE

# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# archive
# ------------------------------------------------------------------------------------------
@app.route('/archive', methods =['GET'], defaults={'req_path': ''})
@app.route('/archive/<path:req_path>')
def archive(req_path):
    if not session.get('has_login', False): return redirect(url_for('login'))
    if not 'A' in session['admind']: return redirect(url_for('upload'))
    abs_path = os.path.join(app.config['archives'], req_path) # Joining the base and the requested path
    #if req_path:print(f"◦ {session['uid']} trying to download {req_path}")
    if not os.path.exists(abs_path): 
        dprint(f"⇒ requested file was not found {abs_path}") #Return 404 if path doesn't exist
        return abort(404) 
    if os.path.isfile(abs_path):  #print(f"◦ sending file ")
        dprint(f'● {session["uid"]} ◦ {session["named"]} just downloaded the file {req_path}')
        return send_file(abs_path) # Check if path is a file and serve
    return render_template('archive.html')
# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# download
# ------------------------------------------------------------------------------------------
@app.route('/download', methods =['GET'], defaults={'req_path': ''})
@app.route('/download/<path:req_path>')
def download(req_path):
    if not session.get('has_login', False): return redirect(url_for('login'))
    if 'D' not in session['admind']:  return redirect(url_for('upload'))
    abs_path = os.path.join(app.config['downloads'], req_path) # Joining the base and the requested path
    #if req_path:print(f"◦ {session['uid']} trying to download {req_path}")
    if not os.path.exists(abs_path): 
        dprint(f"⇒ requested file was not found {abs_path}") #Return 404 if path doesn't exist
        return abort(404) # print(f"◦ requested file was not found") #Return 404 if path doesn't exist
    if os.path.isfile(abs_path):  #print(f"◦ sending file ")
        dprint(f'● {session["uid"]} ◦ {session["named"]} just downloaded the file {req_path}')
        return send_file(abs_path) # Check if path is a file and serve
    return render_template('download.html')
# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# upload
# ------------------------------------------------------------------------------------------
@app.route('/upload', methods =['GET', 'POST'])
def upload():
    if not session.get('has_login', False): return redirect(url_for('login'))
    #print( session['admind'])
    form = UploadFileForm()
    folder_name = os.path.join( app.config['uploads'], session['uid']) 
    #print(f"user {session['uid']} landed on upload page")    
    #print(f"..... has uploaded {len(file_list)} items")
    
    if form.validate_on_submit() and ('U' in session['admind']):
        dprint(f"⇒ user {session['uid']} ◦ {session['named']} is trying to upload {len(form.file.data)} items.")
        if app.config['muc']==0: 
            return render_template('upload.html', form=form, status=[(0, f'✗ Uploads are disabled')])
        else:
            result = []
            n_success = 0
            #---------------------------------------------------------------------------------
            for file in form.file.data:
                isvalid, sf = VALIDATE_FILENAME(secure_filename(file.filename))
                #print(f"[...........] {file.filename} {sf}")
            #---------------------------------------------------------------------------------
                
                if not isvalid:
                    why_failed =  f"✗ File not accepted [{sf}] " if REQUIRED_FILES else f"✗ Extension is invalid [{sf}] "
                    result.append((0, why_failed))
                    continue

                file_name = os.path.join(folder_name, sf)
                if not os.path.exists(file_name):
                    #file_list = os.listdir(folder_name)
                    if len(session['filed'])>=app.config['muc']:
                        why_failed = f"✗ Upload limit reached [{sf}] "
                        result.append((0, why_failed))
                        continue

                file.save(file_name) 
                why_failed = f"✓ Uploaded new file [{sf}] "
                result.append((1, why_failed))
                n_success+=1
                if sf not in session['filed']: session['filed'] = session['filed'] + [sf]

        #---------------------------------------------------------------------------------
            
            #print(f"◦ upload results: \n{result}")
            result_show = ''.join([f'\t{r[-1]}\n' for r in result])
            dprint(f'✓ {session["uid"]} ◦ {session["named"]} just uploaded {n_success} file(s)\n\n{result_show}') 
            return render_template('upload.html', form=form, status=result)
        
    #file_list = session['filed'] #os.listdir(folder_name)
    return render_template('upload.html', form=form, status=(INITIAL_UPLOAD_STATUS if app.config['muc']!=0 else [(-1, f'Uploads are disabled')]))
# ------------------------------------------------------------------------------------------

@app.route('/uploadf', methods =['GET'])
def uploadf():
    r""" force upload - i.e., refresh by using os.list dir """
    if not session.get('has_login', False): return redirect(url_for('login'))
    folder_name = os.path.join( app.config['uploads'], session['uid']) 
    session['filed'] = os.listdir(folder_name)
    return redirect(url_for('upload'))

# ------------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# purge
# ------------------------------------------------------------------------------------------
@app.route('/purge', methods =['GET'])
def purge():
    r""" purges all files that a user has uploaded in their respective uplaod directory
    NOTE: each user will have its won directory, so choose usernames such that a corresponding folder name is a valid one
    """
    if not session.get('has_login', False): return redirect(url_for('login'))
    if 'U' not in session['admind']:  return redirect(url_for('upload'))
    folder_name = os.path.join( app.config['uploads'], session['uid']) 
    if os.path.exists(folder_name):
        file_list = os.listdir(folder_name)
        for f in file_list: os.remove(os.path.join(folder_name, f))
        #print(f"◦ {session['uid']} has purged their files.")
        dprint(f'● {session["uid"]} ◦ {session["named"]} used purge')
        session['filed']=[]
        #dprint(f"filed @ purge= {session['filed']}")
    return redirect(url_for('upload'))
# ------------------------------------------------------------------------------------------

 
# ------------------------------------------------------------------------------------------
# administrative
# ------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
@app.route('/admin/', methods =['GET'], defaults={'req_cmd': ''})
@app.route('/admin/<req_cmd>')
def adminpage(req_cmd):
    r""" opens admin page """ 
    if not session.get('has_login', False): return redirect(url_for('login')) # "Not Allowed - Requires Login"
    if '+' in session['admind']: 
        in_cmd = f'{req_cmd}'
        if in_cmd: 
            if   in_cmd=="upd": STATUS, SUCCESS = update_dl()
            elif in_cmd=="upa": STATUS, SUCCESS = update_al()
            elif in_cmd=="dbw": STATUS, SUCCESS = persist_db()
            elif in_cmd=="dbr": STATUS, SUCCESS = reload_db()
            elif in_cmd=="ref": STATUS, SUCCESS = refresh_board()
            else: STATUS, SUCCESS =  f"Invalid command '{in_cmd}'", False
        else: STATUS, SUCCESS =  f"Admin Access", True
    else: STATUS, SUCCESS =  "This action requires admin privilege", False
    return render_template('admin.html',  status=STATUS, success=SUCCESS)

def update_dl():
    r""" refreshes the  downloads"""
    app.config['dfl'] = GET_FILE_LIST(DOWNLOAD_FOLDER_PATH)
    dprint(f"▶ {session['uid']} ◦ {session['named']} just refreshed the download list.")
    return "Updated download-list", True #  STATUS, SUCCESS

def update_al():
    r""" refreshes the  downloads"""
    app.config['afl'] = GET_FILE_LIST(ARCHIVE_FOLDER_PATH)
    dprint(f"▶ {session['uid']} ◦ {session['named']} just refreshed the archive list.")
    return "Updated archive-list", True #  STATUS, SUCCESS

def persist_db():
    r""" writes db to disk """
    global db
    if write_db_to_disk(db):
        dprint(f"▶ {session['uid']} ◦ {session['named']} just persisted the db to disk.")
        STATUS, SUCCESS = "Persisted db to disk", True
    else: STATUS, SUCCESS =  f"Write error '{args.login}' might be open", False
    return STATUS, SUCCESS 

def reload_db():
    r""" reloads db from disk """
    global db#, HAS_PENDING
    db = read_db_from_disk()
    #HAS_PENDING=0
    dprint(f"▶ {session['uid']} ◦ {session['named']} just reloaded the db from disk.")
    return "Reloaded db from disk", True #  STATUS, SUCCESS

def refresh_board():
    r""" refreshes the  board"""
    if update_board():
        dprint(f"▶ {session['uid']} ◦ {session['named']} just refreshed the board.")
        return "Board was refreshed", True
    else: return "Board not enabled", False


# ------------------------------------------------------------------------------------------
# password reset
# ------------------------------------------------------------------------------------------
@app.route('/x/', methods =['GET'], defaults={'req_uid': ''})
@app.route('/x/<req_uid>')
def repass(req_uid):
    r""" reset user password"""
    if not session.get('has_login', False): return redirect(url_for('login')) # "Not Allowed - Requires Login"
    if '+' in session['admind']: 
        in_uid = f'{req_uid}'
        if in_uid: 
            in_query = in_uid if not args.case else (in_uid.upper() if args.case>0 else in_uid.lower())
            global db#, HAS_PENDING
            record = db.get(in_query, None)
            #print(f"◦ record matched? [{record is not None}]")
            if record is not None: 
                admind, uid, named, _ = record
                admind = f'{admind}'.upper()
                if '+' not in admind:
                    db[uid][3]='' ## 3 for PASS  record['PASS'].values[0]=''
                    #HAS_PENDING+=1
                    dprint(f"▶ {session['uid']} ◦ {session['named']} just reset the password for {uid} ◦ {named}")
                    STATUS, SUCCESS =  f"Password was reset for {uid} {named}", True
                else: STATUS, SUCCESS =  f"Cannot reset password for admin account '{in_query}'", False
            else: STATUS, SUCCESS =  f"User '{in_query}' not found - cannot reset password", False
        else: STATUS, SUCCESS =  f"Requires uid - use /x/<uid> to reset password", False
    else: STATUS, SUCCESS =  "This action requires admin privilege", False
    return render_template('admin.html',  status=STATUS, success=SUCCESS)
# ------------------------------------------------------------------------------------------













#%% @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#<-------------------DO NOT WRITE ANY CODE AFTER THIS
start_time = datetime.datetime.now()
print('◉ start server @ [{}]'.format(start_time))
endpoint = f'{args.host}:{args.port}' if args.host!='0.0.0.0' else f'localhost:{args.port}'
print(f'◉ http://{endpoint}')
serve(app, # https://docs.pylonsproject.org/projects/waitress/en/stable/runner.html
    host = args.host,          
    port = args.port,          
    url_scheme = 'http',     
    threads = args.threads,    
    connection_limit = args.maxconnect,
    max_request_body_size = MAX_UPLOAD_SIZE,
)
#<-------------------DO NOT WRITE ANY CODE AFTER THIS
end_time = datetime.datetime.now()
print('◉ stop server @ [{}]'.format(end_time))

final_db_persist = write_db_to_disk(db)
print('↷ persisted db [{}]'.format(final_db_persist))

if bool(parsed.coe):
    print(f'↪ Cleaning up html/css templates...')
    try:
        for k,v in HTML_TEMPLATES.items():
            h = os.path.join(TEMPLATES_DIR, f"{k}.html")
            if  os.path.isfile(h) : os.remove(h)
        print(f'↪ Creating css templates @ {STATIC_DIR}')
        for k,v in CSS_TEMPLATES.items():
            h = os.path.join(STATIC_DIR, f"{k}.css")
            if os.path.isfile(h): os.remove(h)
        os.removedirs(TEMPLATES_DIR)
        os.removedirs(STATIC_DIR)
        print(f'↪ Removed html/css templates @ {PYDIR}')
    except:
        print(f'↪ Could not remove html/css templates @ {PYDIR}')
print('◉ server up-time was [{}]'.format(end_time - start_time))
print(f'{g_break}\n...Finished!\n{g_break}')



#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__doc__=r"""
:py:mod:`known/basic.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__all__ = [  'Kio', 'Verbose', 'Remap',  'IndexedDict', 'Zipper', 'Mailer', 'BaseConvert' ]
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
from typing import Any, Union, Iterable, Callable #, BinaryIO, cast, Dict, Optional, Type, Tuple, IO
import os, platform, datetime, smtplib, mimetypes, json, pickle
from math import floor, log, ceil
from zipfile import ZipFile
from email.message import EmailMessage
from collections import UserDict
from io import BytesIO
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Kio:
    r""" provides input/out methods for loading saving python objects using json and pickle """
    IOAS = dict(json=json, pickle=pickle)
    IOFLAG = dict(json='', pickle='b')

    @staticmethod
    def save_buffer(o:Any, ioas:str, seek0=False, **kwargs) -> None:
        d = __class__.IOAS.get(f'{ioas}', None)
        assert d is not None, f'key error {ioas}'
        buffer = BytesIO()
        d.dump(o, buffer)
        if seek0: buffer.seek(0) # prepares for reading
        return buffer

    @staticmethod
    def load_buffer(buffer:BytesIO, ioas:str, seek0=True): 
        d = __class__.IOAS.get(f'{ioas}', None)
        assert d is not None, f'key error {ioas}'
        if seek0: buffer.seek(0) # prepares for reading
        return d.load(buffer)

    @staticmethod
    def save_file(o:Any, path:str, ioas:str, **kwargs) -> None:
        d = __class__.IOAS.get(f'{ioas}', None)
        assert d is not None, f'key error {ioas}'
        with open(path, f'w{__class__.IOFLAG[d]}') as f: d.dump(o, f, **kwargs)
        return path

    @staticmethod
    def load_file(path:str, ioas:str):
        d = __class__.IOAS.get(f'{ioas}', None)
        assert d is not None, f'key error {ioas}'
        with open(path, f'r{__class__.IOFLAG[d]}') as f: o = d.load(f)
        return o

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=


class Verbose:
    r""" Contains shorthand helper functions for printing outputs and representing objects as strings.

    .. note::
        This class contains only static methods.
    """
    class Symbols:
        CORRECT =       '✓'
        INCORRECT =     '✗'
        ALPHA =         'α'
        BETA =          'β'
        GAMMA =         'γ'
        DELTA =         'δ'
        EPSILON =       'ε'
        ZETA =          'ζ'
        ETA =           'η'
        THETA =         'θ'
        KAPPA =         'κ'
        LAMBDA =        'λ'
        MU =            'μ' 
        XI =            'ξ'
        PI =            'π'
        ROH =           'ρ'
        SIGMA =         'σ'
        PHI =           'φ'
        PSI =           'Ψ'
        TAU =           'τ'
        OMEGA =         'Ω'
        TRI =           'Δ'
    
    DEFAULT_DATE_FORMAT = ["%Y","%m","%d","%H","%M","%S","%f"]
    r""" Default date format for :func:`~known.basic.Verbose.strU` """

    DASHED_LINE = "=-=-=-=-==-=-=-=-="
    DOCSTR_FORM = lambda x: f'\t!docstr:\n! - - - - - - - - - - - - - - - - -\n{x}\n- - - - - - - - - - - - - - - - - !'

    @staticmethod
    def strN(s:str, n:int) -> str:  
        r""" Repeates a string n-times """
        return ''.join([s for _ in range(n)])

    @staticmethod
    def _recP_(a, level, index, pindex, tabchar='\t', show_dim=False):
        # helper function for recP - do not use directly
        if index<0: index=''
        dimstr = ('* ' if level<1 else f'*{level-1} ') if show_dim else ''
        pindex = f'{pindex}{index}'
        if len(a.shape)==0:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex}\t {a} ]') 
        else:
            print(f'{__class__.strN(tabchar, level)}[ {dimstr}@{pindex} #{a.shape[0]}')
            for i,s in enumerate(a):
                __class__._recP_(s, level+1, i, pindex, tabchar, show_dim)
            print(f'{__class__.strN(tabchar, level)}]')

    @staticmethod
    def recP(arr:Iterable, show_dim:bool=False) -> None: 
        r"""
        Recursive Print - print an iterable recursively with added indentation.

        :param arr:         any iterable with ``shape`` property.
        :param show_dim:    if `True`, prints the dimension at the start of each item
        """
        __class__._recP_(arr, 0, -1, '', '\t', show_dim)
    
    @staticmethod
    def strA(arr:Iterable, start:str="", sep:str="|", end:str="") -> str:
        r"""
        String Array - returns a string representation of an iterable for printing.
        
        :param arr:     input iterable
        :param start:   string prefix
        :param sep:     item seperator
        :param end:     string postfix
        """
        res=start
        for a in arr: res += (str(a) + sep)
        return res + end

    @staticmethod
    def strA_(arr:Iterable, start:str="", sep:str="|", end:str="") -> None: print(__class__.strA(arr, start, sep, end))
    
    @staticmethod
    def strD(arr:Iterable, sep:str="\n", cep:str=":\n", caption:str="") -> str:
        r"""
        String Dict - returns a string representation of a dict object for printing.
        
        :param arr:     input dict
        :param sep:     item seperator
        :param cep:     key-value seperator
        :param caption: heading at the top
        """
        res=f"=-=-=-=-==-=-=-=-={sep}DICT #[{len(arr)}] : {caption}{sep}{__class__.DASHED_LINE}{sep}"
        for k,v in arr.items(): res+=str(k) + cep + str(v) + sep
        return f"{res}{__class__.DASHED_LINE}{sep}"

    @staticmethod
    def strD_(arr:Iterable, sep:str="\n", cep:str=":\n", caption:str="") -> None: __class__.strD(arr, sep, cep, caption)

    @staticmethod
    def strU(form:Union[None, Iterable[str]], start:str='', sep:str='', end:str='') -> str:
        r""" 
        String UID - returns a formated string of current timestamp.

        :param form: the format of timestamp, If `None`, uses the default :data:`~known.basic.Verbose.DEFAULT_DATE_FORMAT`.
            Can be selected from a sub-set of ``["%Y","%m","%d","%H","%M","%S","%f"]``.
            
        :param start: UID prefix
        :param sep: UID seperator
        :param end: UID postfix

        .. seealso::
            :func:`~known.basic.uid`
        """
        if not form: form = __class__.DEFAULT_DATE_FORMAT
        return start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end

    @staticmethod
    def now(year:bool=True, month:bool=True, day:bool=True, 
            hour:bool=True, minute:bool=True, second:bool=True, mirco:bool=True, 
            start:str='', sep:str='', end:str='') -> str:
        r""" Unique Identifier - useful in generating unique identifiers based on current timestamp. 
        Helpful in generating unique filenames based on timestamps. 
        
        .. seealso::
            :func:`~known.basic.Verbose.strU`
        """
        form = []
        if year:    form.append("%Y")
        if month:   form.append("%m")
        if day:     form.append("%d")
        if hour:    form.append("%H")
        if minute:  form.append("%M")
        if second:  form.append("%S")
        if mirco:   form.append("%f")
        assert (form), 'format should not be empty!'
        return (start + datetime.datetime.strftime(datetime.datetime.now(), sep.join(form)) + end)

    @staticmethod
    def show(x:Any, cep:str='\t\t:', sw:str='__', ew:str='__') -> None:
        r"""
        Show Object - describes members of an object using the ``dir`` call.

        :param x:       the object to be described
        :param cep:     the name-value seperator
        :param sw:      argument for ``startswith`` to check in member name
        :param ew:      argument for ``endswith`` to check in member name

        .. note:: ``string.startswith`` and ``string.endswith`` checks are performed on each member of the object 
            and only matching member are displayed. This is usually done to prevent showing dunder members.
        
        .. seealso::
            :func:`~known.basic.Verbose.showX`
        """
        for d in dir(x):
            if not (d.startswith(sw) or d.endswith(ew)):
                v = ""
                try:
                    v = getattr(x, d)
                except:
                    v='?'
                print(d, cep, v)

    @staticmethod
    def showX(x:Any, cep:str='\t\t:') -> None:
        """ Show Object (Xtended) - describes members of an object using the ``dir`` call.

        :param x:       the object to be described
        :param cep:     the name-value seperator

        .. note:: This is the same as :func:`~known.basic.Verbose.show` but skips ``startswith`` and ``endswith`` checks,
            all members are shown including dunder members.

        .. seealso::
            :func:`~known.basic.Verbose.show`
        """
        for d in dir(x):
            v = ""
            try:
                v = getattr(x, d)
            except:
                v='?'
            print(d, cep, v)

    @staticmethod
    def dir(x:Any, doc=False, filter:str='', sew=('__','__')):
        """ Calls ```dir``` on given argument and lists the name and types of non-dunder members.

        :param filter: csv string of types to filter out like `type,function,module`, keep blank for no filter
        :param doc: shows docstring ```__doc``` 
            If ```doc``` is True, show all member's ```__doc__```.
            If ```doc``` is False, does not show any ```__doc__```. 
            If ```doc``` is a string, show ```__doc__``` of specific types only given by csv string.

        :param sew: 2-Tuple (start:str, end:str) - excludes member names that start and end with specific chars, 
            used to exclude dunder methods by default
        """
        #if self_doc: print( f'{type(x)}\n{x.__doc__}\n' )
        if sew: sw, ew = f'{sew[0]}', f'{sew[1]}'
        doc_is_specified = (isinstance(doc, str) and bool(doc))
        if doc_is_specified: doc_match =[ t for t in doc.replace(' ','').split(',') if t ]
        if filter: filter_match =[ t for t in filter.replace(' ','').split(',') if t ]
        counter=1
        for k in dir(x):
            if sew:
                if (k.startswith(sw) and k.endswith(ew)): continue
            m = getattr(x,k)
            n = str(type(m)).split("'")[1]
            if filter:
                if not (n in filter_match):  continue
            s = f'[{counter}] {k} :: {n}'#.encode('utf-16')

            if doc:
                if doc_is_specified:
                    if n in doc_match: 
                        d = __class__.DOCSTR_FORM(m.__doc__)
                    else:
                        d=''
                else:
                    d = __class__.DOCSTR_FORM(m.__doc__)
            else:
                d = ''
            counter+=1
            print(f'{s}{d}')


    @staticmethod
    def info(x:Any, show_object:bool=False):
        r""" Shows the `type`, `length` and `shape` of an object and optionally shows the object as well.

        :param x:           the object to get info about
        :param show_object: if `True`, prints the object itself

        .. note:: This is used to check output of some functions without having to print the full output
            which may take up a lot of console space. Useful when the object are of nested types.

        .. seealso::
            :func:`~known.basic.Verbose.infos`
        """
        print(f'type: {type(x)}')
        if hasattr(x, '__len__'):
            print(f'len: {len(x)}')
        if hasattr(x, 'shape'):
            print(f'shape: {x.shape}')
        if show_object:
            print(f'object:\n{x}')

    @staticmethod
    def infos(X:Iterable, show_object=False):
        r""" Shows the `type`, `length` and `shape` of each object in an iterable 
        and optionally shows the object as well.

        :param x:           the object to get info about
        :param show_object: if `True`, prints the object itself

        .. seealso::
            :func:`~known.basic.Verbose.info`
        """
        for t,x in enumerate(X):
            print(f'[# {t}]')
            __class__.info(x, show_object=show_object)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Remap:
    r""" 
    Provides a mapping between ranges, works with scalars, ndarrays and tensors.

    :param Input_Range:     *FROM* range for ``forward`` call, *TO* range for ``backward`` call
    :param Output_Range:    *TO* range for ``forward`` call, *FROM* range for ``forward`` call
    """

    def __init__(self, Input_Range:tuple, Output_Range:tuple) -> None:
        r"""
        :param Input_Range:     `from` range for ``i2o`` call, `to` range for ``o2i`` call
        :param Output_Range:    `to` range for ``i2o`` call, `from` range for ``o2i`` call
        """
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

    def set_input_range(self, Range:tuple) -> None:
        r""" set the input range """
        self.input_low, self.input_high = Range
        self.input_delta = self.input_high - self.input_low

    def set_output_range(self, Range:tuple) -> None:
        r""" set the output range """
        self.output_low, self.output_high = Range
        self.output_delta = self.output_high - self.output_low

    def backward(self, X):
        r""" maps ``X`` from ``Output_Range`` to ``Input_Range`` """
        return ((X - self.output_low)*self.input_delta/self.output_delta) + self.input_low

    def forward(self, X):
        r""" maps ``X`` from ``Input_Range`` to ``Output_Range`` """
        return ((X - self.input_low)*self.output_delta/self.input_delta) + self.output_low

    def __call__(self, X, backward=False):
        return self.backward(X) if backward else self.forward(X)
    
    def swap_range(self):
        Input_Range, Output_Range = (self.output_low, self.output_high), (self.input_low, self.input_high)
        self.set_input_range(Input_Range)
        self.set_output_range(Output_Range)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class BaseConvert:
    
    r""" Number System Conversion 
    
    A number is abstract concept that has many representations using sets of symbols

    A base-n number system uses a set of n digits to represent any number
    This is called the representation of the number

    Given one representation, we only need to convert to another

    """

    @staticmethod
    def zeros(n): return [0 for _ in range(n)]

    @staticmethod
    def convert(digits, base_from, base_to, reversed=True):
        r""" convers from one base to another 
        
        :param digits:      iterable of digits in base ```base_from```. NOTE: digits are Natural Numbers starting at 0. base 'b' will have digits between [0, b-1]
        :param base_from:   int - the base to convert from
        :param base_to:     int - the base to convert to
        :param reversed:    bool - if True, digits are assumed in reverse (human readable left to right)
                            e.g. if reversed is True then binary digits iterable [1,0,0] will represent [4] in decimal otherwise it will represent [1] in decimal
        """

        digits_from =  [int(abs(d)) for d in digits] # convert to int data-type
        if reversed: digits_from = digits_from[::-1]
        ndigits_from = len(digits_from)
        mult_from = [base_from**i for i in range(ndigits_from)]
        repr_from = sum([ui*vi for ui,vi in zip(digits_from,mult_from, strict=True)]) #dot(digits_from , mult_from)

        #ndc = base_from**ndigits_from
        ndigits_to = ceil(log(repr_from,base_to))
        digits_to =  __class__.zeros(ndigits_to) 
        n = int(repr_from)
        for d in range(ndigits_to):
            digits_to[d] = n%base_to
            n=n//base_to

        if reversed: digits_to = digits_to[::-1]
        return tuple(digits_to)


    @staticmethod
    def ndigits(num:int, base:int): return ceil(log(num,base))

    @staticmethod
    def int2base(num:int, base:int, digs:int) -> list:
        r""" 
        Convert base-10 integer to a base-n list of fixed no. of digits 

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        :param digs:    no of digits in the output

        :returns:       represented number as a list of ordinals in base-n number system

        .. seealso::
            :func:`~known.basic.base2int`
        """
        
        ndigits = digs if digs else ceil(log(num,base)) 
        digits =  __class__.zeros(ndigits)
        n = num
        for d in range(ndigits):
            digits[d] = n%base
            n=n//base
        return digits

    @staticmethod
    def base2int(num:Iterable, base:int) -> int:
        """ 
        Convert an iterbale of digits in base-n system to base-10 integer

        :param num:     iterable of base-n digits
        :param base:    base-n number system

        :returns:       represented number as a integer in base-10 number system

        .. seealso::
            :func:`~known.basic.int2base`
        """
        res = 0
        for i,n in enumerate(num): res+=(base**i)*n
        return int(res)


    SYM_BIN = { f'{i}':i for i in range(2) }
    SYM_OCT = { f'{i}':i for i in range(8) }
    SYM_DEC = { f'{i}':i for i in range(10) }
    SYM_HEX = {**SYM_DEC , **{ s:(i+10) for i,s in enumerate(('A', 'B', 'C', 'D', 'E', 'F'))}}
    
    @staticmethod
    def n_syms(n): return { f'{i}':i for i in range(n) }

    @staticmethod
    def to_base_10(syms:dict, num:str):
        b = len(syms)
        l = [ syms[n] for n in num[::-1] ]
        return __class__.base2int(l, b)

    @staticmethod
    def from_base_10(syms:dict, num:int, joiner='', ndigs=None):
        base = len(syms)
        #print(f'----{num=} {type(num)}, {base=}, {type(base)}')
        if not ndigs: ndigs = (1 + (0 if num==0 else floor(log(num, base))))  # __class__.ndigs(num, base)
        ss = tuple(syms.keys())
        S = [ ss[i]  for i in __class__.int2base(num, base, ndigs) ]
        return joiner.join(S[::-1])


    @staticmethod
    def int2hex(num:int, joiner=''): return __class__.from_base_10(__class__.SYM_HEX, num, joiner)
  
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class IndexedDict(UserDict):
    r""" Implements an Indexed dict where values can be addressed using both index(int) and keys(str) """

    def __init__(self, **members) -> None:
        self.names = []
        super().__init__(*[], **members)
    
    def keys(self): return enumerate(self.names, 0) # for i,k in self.keys()

    def items(self): return enumerate(self.data.items(), 0) # for i,(k,v) in self.items()

    def __len__(self): return len(self.data)

    def __getitem__(self, name): 
        if isinstance(name, int): name = self.names[name]
        if name in self.data: 
            return self.data[name]
        else:
            raise KeyError(name)

    def __setitem__(self, name, item): 
        if isinstance(name, int): name = self.names[name]
        if name not in self.data: self.names.append(name)
        self.data[name] = item

    def __delitem__(self, name): 
        index = None
        if isinstance(name, int):  
            index = name
            name = self.names[name]
        if name in self.data: 
            del self.names[self.names.index(name) if index is None else index]
            del self.data[name]

    def __iter__(self): return iter(self.names)

    def __contains__(self, name): return name in self.data

    # Now, add the methods in dicts but not in MutableMapping

    def __repr__(self) -> str:
        return f'{__class__} :: {len(self)} Members'
    
    def __str__(self) -> str:
        items = ''
        for i,k in enumerate(self):
            items += f'[{i}] \t {k} : {self[i]}\n'
        return f'{__class__} :: {len(self)} Members\n{items}'
    
    def __copy__(self):
        inst = self.__class__.__new__(self.__class__)
        inst.__dict__.update(self.__dict__)
        # Create a copy and avoid triggering descriptors
        inst.__dict__["data"] = self.__dict__["data"].copy()
        inst.__dict__["names"] = self.__dict__["names"].copy()
        return inst

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Zipper:
    r""" zip API using ZipFile package """

    @staticmethod
    def zip_files(zip_path:str, files, **kwargs):
        r""" zips all (only files) in the list of file paths and saves at 'zip_path' """
        zipped = 0
        if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'
        with ZipFile(zip_path, 'w', **kwargs) as zip_object:
            for path in files:
                if not os.path.isfile(path): continue
                zip_object.write(f'{path}')
                zipped+=1
        return zipped, zip_path

    @staticmethod
    def get_all_file_paths(directory):
        r""" recursively list all files in a folder """
        file_paths = []
        # crawling through directory and subdirectories
        for root, directories, files in os.walk(directory):
            for filename in files:
                # join the two strings in order to form the full filepath.
                filepath = os.path.join(root, filename)
                file_paths.append(filepath)
        return file_paths   

    @staticmethod
    def zip_folders(zip_path:str, folders, **kwargs):  
        r""" zip multiple folders into a single zip file """    
        if isinstance(folders, str): folders= [f'{folders}']

        if not zip_path : zip_path = f'{folders[0]}.zip'
        if not zip_path.lower().endswith('.zip'): zip_path = f'{zip_path}.zip'  
        all_files = []
        for folder in folders: all_files.extend(__class__.get_all_file_paths(folder))
        return __class__.zip_files(f'{zip_path}', all_files, **kwargs)
    
    @staticmethod
    def zip_folder(folder:str, **kwargs):
        r""" zip a single folder with the same zip file name """     
        return  __class__.zip_files(f'{folder}.zip', __class__.get_all_file_paths(folder),  **kwargs)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Mailer:
    r""" Use a g-mail account to send mail. 

    .. warning:: security concern
        You should enable 2-factor-auth in gmail and generate an app-password instead of using your gmail password.
        Visit (https://myaccount.google.com/apppasswords) to generate app-password.
        Usually, these type of emails are treated as spam by google, so they must be marked 'not spam' at least once.
        It is recomended to create a seperate gmail account for sending mails.
    """
    
    DEFAULT_CTYPE = 'application/octet-stream'  

    @staticmethod
    def global_alias(prefix=''): return f'{prefix}{os.getlogin()} @ {platform.node()}:{platform.system()}.{platform.release()}'

    @staticmethod
    def get_mime_types(files):
        r""" gets mimetype info all files in a list """
        if isinstance(files, str): files=[f'{files}']
        res = []
        for path in files:
            if not os.path.isfile(path): continue
            ctype, encoding = mimetypes.guess_type(path)
            if ctype is None or encoding is not None: ctype = __class__.DEFAULT_CTYPE  
            maintype, subtype = ctype.split('/', 1)
            res.append( (path, maintype, subtype) )
        return res

    @staticmethod
    def compose_mail( subject:str, rx:str, cc:str, bcc:str, content:str, signature:str, attached, verbose=True):
        r""" compose an e-mail msg to send later
        
        :param subject: subject
        :param rx: csv recipent email address
        :param cc: csv cc email address
        :param content: main content
        :param attached: list of attached files - is a 2-tupe (attachment_type, (args...) )

        # attach all files in the list :: ('', ('file1.xyz', 'file2.xyz'))
        # zip all the files in the list :: ('zipname.zip', '(file1.xyz', 'file2.xyz'))
        """
        
        msg = EmailMessage()

        # set subject
        msg['Subject'] = f'{subject}'
        if verbose: print(f'SUBJECT: {subject}')

        # set to
        msg['To'] = rx
        if verbose: print(f'TO: {rx}')

        if cc: msg['Cc'] = cc
        if verbose: print(f'CC: {cc}')

        if bcc: msg['Bcc'] = bcc
        if verbose: print(f'BCC: {bcc}')

        # set content
        body = content + signature
        msg.set_content(body)
        if verbose: print(f'MESSAGE: #[{len(body)}] chars.')

        default_attached = []

        attached = [] if attached is None else attached
        for (attach_type, attach_args) in attached:
            if verbose: print(f' ... processing attachement :: {attach_type} :: {attach_args}')

            all_files = []
            for path in attach_args:
                if os.path.isdir(path):
                    all_files.extend(Zipper.get_all_file_paths(path))
                elif os.path.isfile(path):
                    all_files.append(path)
                else:
                    if verbose: print(f'[!] Invalid Path :: {path}, skipped...')

            if not attach_type:  # attach individually
                default_attached.extend(__class__.get_mime_types(all_files))
            else: # make zip
                zipped, zip_path=Zipper.zip_files(attach_type, all_files)
                if verbose: print(f'\t --> zipped {zipped} items @ {zip_path} ')
                if zipped>0:
                    default_attached.extend(__class__.get_mime_types(zip_path))
                else:
                    if verbose: print(f'[!] [{zip_path}] is empty, will not be attched!' )
                    try:
                        os.remove(zip_path)
                        if verbose: print(f'[!] [{zip_path}] was removed.' )
                    except:
                        if verbose: print(f'[!] [{zip_path}] could not be removed.' ) 
                

        # set attached ( name, main_type, sub_type), if sub_type is none, auto-infers using imghdr
        for file_name,main_type,sub_type in default_attached:
            if verbose: print(f'[+] Attaching file [{main_type}/{sub_type}] :: [{file_name}]')
            with open (file_name, 'rb') as f: 
                file_data = f.read()
            msg.add_attachment(file_data, maintype=main_type, subtype=sub_type, filename=os.path.basename(file_name))

        return msg

    @staticmethod
    def send_mail(login, msg, verbose=True):
        r""" send a msg using smtp.gmail.com:587 with provided credentials """
        username, password = login()
        if verbose: print(f'[*] Sending Email from {username}')
        msg['From'] = f'{username}' # set from
        with smtplib.SMTP('smtp.gmail.com', 587) as smpt:
            smpt.starttls()
            smpt.login(username, password)
            smpt.ehlo()
            smpt.send_message(msg)
        if verbose: print(f'[*] Sent!')

    @staticmethod
    def send(login:Union[Callable, tuple, list, dict], subject:str, rx:str, cc:str, bcc:str, content:str, signature:str, attached, verbose=True):
        if isinstance(login, (list, tuple)):
            username, password = login
            login = lambda: (username, password)
        elif isinstance(login, dict):
            username, password = login.get('username'), login.get('password')
            login = lambda: (username, password)
        else:   pass # assume callable
        __class__.send_mail(login,  __class__.compose_mail(subject, rx, cc, bcc, content, signature, attached, verbose), verbose)

        
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=




import pika


class Client:
    def __init__(self, server, queue='default'): 
        self.server, self.queue = server, queue
    
    def connect(self):
        self.connection = pika.BlockingConnection(pika.ConnectionParameters(host=self.server))
        self.channel = self.connection.channel()
        self.channel.queue_declare(queue=self.queue)

    def send_msg(self, msg):
        # publish on the queue
        self.channel.basic_publish(
        exchange='', 
        routing_key=self.queue, 
        body=msg,
        )

    def send_buffer(self, buf):
        # publish on the queue
        buf.seek(0)
        self.channel.basic_publish(
        exchange='', 
        routing_key=self.queue, 
        body=buf.read(),
        )

    def disconnect(self): self.connection.close()




import requests

class Client:
    r""" Client class for making api calls (http reuqests)

    :param server: <str> the address of api server
    :param port:   <str> the port of api server


    To send json data, use: Client.send_json(json_object)
    ---> The <json_object> can be a dict, list or string

    To send buffers, use: Client.send_buffer(**buffers):
    ---> The <buffers> are named buffers provided as keyword arguments like {buffer-name : buffer-object}
    ---> If using buffer, make sure to buffer.seek(0) before sending 

    To send files, use: Client.send_file(*handles):
    ---> The <handles> is a list of file names


    The response contains 3 things - msg, mtype, mcode
    msg :   base object sent by server - a string, list or dict
    mtype:  type of msg object - can be 'text', 'json', 'bytes' or 'error'  (error has text type msg)
    mcode:  status code - can be True or False (true when HTTP response was recieved, false otherwise)
    """
    
    class MTypes:
        ERROR = -1
        NONE = 0
        TEXT = 1
        JSON = 2
        BYTES = 3
        _rev = {
            -1 : 'ERROR',
            0 : 'NONE',
            1 : 'TEXT',
            2 : 'JSON',
            3 : 'BYTES',

        }
        def code(c): return __class__._rev.get(int(c), None)

    def __init__(self, server='localhost', port='8080'): 
        self.server = f'http://{server}:{port}/' # http://localhost:8080/
    
    def handle_response(self, response):
        self.response = response
        if self.response.ok:
            ct = self.response.headers['Content-Type']
            if ct.startswith('text'):                       msg, mtype = self.response.text, __class__.MTypes.TEXT
            elif ct.startswith('application/json'):         msg, mtype = self.response.json(), __class__.MTypes.JSON
            else:                                           msg, mtype = self.response.content, __class__.MTypes.BYTES
        else:                                               msg, mtype = (f'[{self.response.status_code}]'), __class__.MTypes.NONE
        return msg, mtype, True


    def send_args(self, **kwargs): 
        try:                    msg, mtype, mcode = self.handle_response(requests.post(url=self.server,data=kwargs))
        except:                 msg, mtype, mcode = f"[SEND_ERROR:ARGS] cannot send ARGS to {self.server}", __class__.MTypes.ERROR, False
        return                  msg, mtype, mcode

    def send_json(self, json_object): 
        try:                    msg, mtype, mcode = self.handle_response(requests.post(url=self.server,json=json_object))
        except:                 msg, mtype, mcode = f"[SEND_ERROR:JSON] cannot send JSON to {self.server}", __class__.MTypes.ERROR, False
        return                  msg, mtype, mcode

    def send_buffers_by_name(self, seek0, buffers:dict):
        try:
            assert (len(buffers)>0) 
            if seek0: 
                for b in buffers.values(): b.seek(0)
            msg, mtype, mcode = self.handle_response(requests.post(url=self.server,files=buffers))
        except AssertionError:  msg, mtype, mcode = f"[SEND_ERROR:BUFFER] NO BUFFERS provided", __class__.MTypes.ERROR, False
        except:                 msg, mtype, mcode = f"[SEND_ERROR:BUFFER] cannot send BUFFERS to {self.server}", __class__.MTypes.ERROR, False
        return                  msg, mtype, mcode


    def send_files_by_handles(self, handles:list): return self.send_buffers_by_name(False, {h.name:h for h in handles})

    def send_files_by_named_handels(self, handles:dict): return self.send_buffers_by_name(False, handles)

    def send_files_by_path(self, paths:list):
        handles = []
        try:
            handles += [open(path, 'r') for path in paths]
            msg, mtype, mcode = self.send_files_by_handles(handles)
        except: msg, mtype, mcode = f"[SEND_ERROR:FILE] Files could not be read", __class__.MTypes.ERROR, False
        finally:
            for h in handles: 
                try:h.close()
                except:pass
        return msg, mtype, mcode

    def send_files_by_named_paths(self, paths:dict): 
        handles = {}
        try:
            handles += {name:open(path, 'r') for name,path in paths.items()}
            msg, mtype, mcode = self.send_files_by_named_handels(handles)
        except: msg, mtype, mcode = f"[SEND_ERROR:FILE] Files could not be read", __class__.MTypes.ERROR, False
        finally:
            for hk, hv in handles.items():
                try: hv.close()
                except: pass
        return   msg, mtype, mcode


    def close(self):
        try: self.response.close()
        except:pass

HelpModule=\
f"""

# global scope functions
# python -m known.api --user=user.py
def handle_args(host, args): return ""
def handle_json(host, data): return ""
def handle_file(host, files): return ""


class UserModuleInstance:
    # creates an instance of this class, init is called
    # python -m known.api --user=user.py --object=UserModuleInstance --callable=1
    def __init__(self): pass
    def handle_args(self, host, args): return ""
    def handle_json(self, host, data): return ""
    def handle_file(self, host, files): return ""

class UserModuleStatic:
    # no instance created, assumed to be a simple object
    # python -m known.api --user=user.py --object=UserModuleStatic --callable=0
    def handle_args(host, args): return ""
    def handle_json(host, data): return ""
    def handle_file(host, files): return ""




"""
# ACCEPTED 		        : 202
# ALREADY_REPORTED 		: 208
# BAD_GATEWAY 		    : 502
# BAD_REQUEST 		    : 400  #<------- used - indicates use of other request type than GET or POST
# CONFLICT 		        : 409
# CONTINUE 		        : 100
# CREATED 		        : 201
# EARLY_HINTS 		    : 103
# EXPECTATION_FAILED 		: 417
# FAILED_DEPENDENCY 		: 424
# FORBIDDEN 		        : 403
# FOUND 		            : 302
# GATEWAY_TIMEOUT 		: 504
# GONE 		            : 410
# HTTP_VERSION_NOT_SUPPORTED 		: 505
# IM_A_TEAPOT 		    : 418
# IM_USED 		        : 226
# INSUFFICIENT_STORAGE 		: 507
# INTERNAL_SERVER_ERROR 		: 500
# LENGTH_REQUIRED 		: 411
# LOCKED 		: 423
# LOOP_DETECTED 		: 508
# METHOD_NOT_ALLOWED 		: 405
# MISDIRECTED_REQUEST 		: 421
# MOVED_PERMANENTLY 		: 301
# MULTIPLE_CHOICES 		: 300
# MULTI_STATUS 		: 207
# NETWORK_AUTHENTICATION_REQUIRED 		: 511
# NON_AUTHORITATIVE_INFORMATION 		: 203
# NOT_ACCEPTABLE 		: 406
# NOT_EXTENDED 		: 510
# NOT_FOUND 		: 404 #<------ used - indicates the return_object was not of correct type (str, dict, list)
# NOT_IMPLEMENTED 		: 501
# NOT_MODIFIED 		: 304
# NO_CONTENT 		: 204
# OK 		: 200
# PARTIAL_CONTENT 		: 206
# PAYMENT_REQUIRED 		: 402
# PERMANENT_REDIRECT 		: 308
# PRECONDITION_FAILED 		: 412
# PRECONDITION_REQUIRED 		: 428
# PROCESSING 		: 102
# PROXY_AUTHENTICATION_REQUIRED 		: 407
# REQUESTED_RANGE_NOT_SATISFIABLE 		: 416
# REQUEST_ENTITY_TOO_LARGE 		: 413
# REQUEST_HEADER_FIELDS_TOO_LARGE 		: 431
# REQUEST_TIMEOUT 		: 408
# REQUEST_URI_TOO_LONG 		: 414
# RESET_CONTENT 		: 205
# SEE_OTHER 		: 303
# SERVICE_UNAVAILABLE 		: 503
# SWITCHING_PROTOCOLS 		: 101
# TEMPORARY_REDIRECT 		: 307
# TOO_EARLY 		: 425
# TOO_MANY_REQUESTS 		: 429
# UNAUTHORIZED 		: 401
# UNAVAILABLE_FOR_LEGAL_REASONS 		: 451
# UNPROCESSABLE_ENTITY 		: 422
# UNSUPPORTED_MEDIA_TYPE 		: 415
# UPGRADE_REQUIRED 		: 426
# USE_PROXY 		: 305
# VARIANT_ALSO_NEGOTIATES 		: 506



# #def query(url, **params): return requests.get(url, params) # works for string key-value pairs only
# #response=query(server, name='Nelson', age='31')
# #vb.show(response)


# # # use any one of data or json
# # response = requests.post(
# #     url=server,
# #     data={'key': 'value'}, # A dictionary, list of tuples, bytes or a file object to send to the specified url
# #     )
# # vb.show(response)


# """

# """ GENERIC METHOD CALLS

# # response = requests.post(
# #     url=server,
# #     data={key: value}, # A dictionary, list of tuples, bytes or a file object to send to the specified url
# #     json={key: value}, # A JSON object to send to the specified url
# #     files={key: value}, # File handles (opened)
# #     timeout=0.0,
# #     )
# # response = requests.get(
# #     url=server,
# #     params={key: value}, # A dictionary, list of tuples, bytes or a file object to send to the specified url
# #     timeout=0.0,
# #     )
# """



__doc__="""
RabbitMQ for Queue based messaging
"""

#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] can not import {__name__}.{__file__}')
#-----------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# imports ----------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
import os, argparse, datetime, sys, importlib, importlib.util

#PYDIR = os.path.dirname(__file__) # script directory of __main__.py
try:
    import pika
except: exit(f'[!] The required packages missing:\tpika\n  ⇒ pip install pika')

class DefaultUserModule:
    def handle_msg(self, body):        print(f"handle_args({body})")

parser = argparse.ArgumentParser()
parser.add_argument('--user', type=str, default='', help="path of main user module file")
parser.add_argument('--object', type=str, default='', help="path of main user module")
parser.add_argument('--callable', type=int, default=0, help="if true, calls the class to create a new instance (with no args)")
parser.add_argument('--handle_msg', type=str, default='handle_msg', help="path of main user module file")
parser.add_argument('--host', type=str, default='', help="host ip-address")
parser.add_argument('--queue', type=str, default='default', help="queue name")
parsed = parser.parse_args()

if not parsed.host: exit(f'Invalid host address')
if not parsed.queue: exit(f'Invalid queue name')
if parsed.user:
    # user-module
    USER_MODULE_FILE_PATH = os.path.abspath(parsed.user)
    print(f'↪ Loading user-module from {USER_MODULE_FILE_PATH}')
    if not os.path.isfile(USER_MODULE_FILE_PATH): exit(f'Invalid user-module file @ {USER_MODULE_FILE_PATH}')
    try: 
        # from https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
        user_module_spec = importlib.util.spec_from_file_location("", USER_MODULE_FILE_PATH)
        user_module = importlib.util.module_from_spec(user_module_spec)
        user_module_spec.loader.exec_module(user_module)
        print(f'↪ Imported user-module from {user_module.__file__}')
    except: exit(f'[!] Could import user-module "{USER_MODULE_FILE_PATH}"')
    if parsed.object:
        try:
            user_module = getattr(user_module, parsed.object)
            if bool(parsed.callable): user_module = user_module()
        except:
            exit(f'Could not load object {parsed.object}')
    USER_HANDLE_MSG = parsed.handle_msg
else:
    print(f'↪ [!] user-module not defined, using default.')
    USER_HANDLE_MSG = 'handle_msg'
    user_module = DefaultUserModule()


if not hasattr(user_module, USER_HANDLE_MSG): exit(f'[!] MSG Handler Method not found {user_module}.{USER_HANDLE_MSG}')
#from known.basic import Verbose as vb
# ------------------------------------------------------------------------------------------
# application setting and instance
# ------------------------------------------------------------------------------------------

# establish connection
try:
    connection = pika.BlockingConnection(pika.ConnectionParameters(host=parsed.host))
    channel = connection.channel()
    channel.queue_declare(queue=parsed.queue)
except: exit(f'Connection cannot be established with {parsed.host}@{parsed.queue}')
def callback(ch, method, properties, body): return getattr(user_module, USER_HANDLE_MSG)(body)
channel.basic_consume(queue=parsed.queue, on_message_callback=callback, auto_ack=True)
start_time = datetime.datetime.now()
print('◉ start server @ [{}]'.format(start_time))
try: channel.start_consuming()
except KeyboardInterrupt:
    try: sys.exit(0)
    except SystemExit: os._exit(0)

#<-------------------DO NOT WRITE ANY CODE AFTER THIS
end_time = datetime.datetime.now()
print('◉ stop server @ [{}]'.format(end_time))
print('◉ server up-time was [{}]'.format(end_time - start_time))





__doc__=r"""
:py:mod:`known/pix.py`
"""
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
__all__ = ['BaseConvertNumpy', 'Pix'] 
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
from typing import Any, Union, Iterable, Callable #, BinaryIO, cast, Dict, Optional, Type, Tuple, IO
from math import floor, log, ceil
import numpy as np
import matplotlib.pyplot as plt
import cv2 # pip install opencv-python


#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class BaseConvertNumpy:
    
    r""" Number System Conversion (using numpy)
    
    A number is abstract concept that has many representations using sets of symbols

    A base-n number system uses a set of n digits to represent any number
    This is called the representation of the number

    Given one representation, we only need to convert to another

    """
    
    @staticmethod
    def convert(digits, base_from, base_to, reversed=True):
        r""" convers from one base to another 
        
        :param digits:      iterable of digits in base ```base_from```. NOTE: digits are Natural Numbers starting at 0. base 'b' will have digits between [0, b-1]
        :param base_from:   int - the base to convert from
        :param base_to:     int - the base to convert to
        :param reversed:    bool - if True, digits are assumed in reverse (human readable left to right)
                            e.g. if reversed is True then binary digits iterable [1,0,0] will represent [4] in decimal otherwise it will represent [1] in decimal
        """

        digits_from =  np.array(digits, dtype=np.uint32) # convert to int data-type
        if reversed: digits_from = digits_from[::-1]
        ndigits_from = len(digits_from)
        mult_from = np.array([base_from**i for i in range(ndigits_from)], dtype=np.uint32)
        repr_from = np.dot(digits_from , mult_from)

        #ndc = base_from**ndigits_from
        ndigits_to = ceil(log(repr_from,base_to))
        digits_to =  np.zeros((ndigits_to,), dtype=np.uint32)
        n = int(repr_from)
        for d in range(ndigits_to):
            digits_to[d] = n%base_to
            n=n//base_to

        if reversed: digits_to = digits_to[::-1]
        return tuple(digits_to)


    @staticmethod
    def ndigits(num:int, base:int): return ceil(log(num,base))

    @staticmethod
    def int2base(num:int, base:int, digs:int) -> list:
        r""" 
        Convert base-10 integer to a base-n list of fixed no. of digits 

        :param num:     base-10 number to be represented
        :param base:    base-n number system
        :param digs:    no of digits in the output

        :returns:       represented number as a list of ordinals in base-n number system

        .. seealso::
            :func:`~known.basic.base2int`
        """
        
        ndigits = digs if digs else ceil(log(num,base)) 
        digits =  np.zeros((ndigits,), dtype=np.uint32)
        n = num
        for d in range(ndigits):
            digits[d] = n%base
            n=n//base
        return digits

    @staticmethod
    def base2int(num:Iterable, base:int) -> int:
        """ 
        Convert an iterbale of digits in base-n system to base-10 integer

        :param num:     iterable of base-n digits
        :param base:    base-n number system

        :returns:       represented number as a integer in base-10 number system

        .. seealso::
            :func:`~known.basic.int2base`
        """
        res = 0
        for i,n in enumerate(num): res+=(base**i)*n
        return int(res)


    SYM_BIN = { f'{i}':i for i in range(2) }
    SYM_OCT = { f'{i}':i for i in range(8) }
    SYM_DEC = { f'{i}':i for i in range(10) }
    SYM_HEX = {**SYM_DEC , **{ s:(i+10) for i,s in enumerate(('A', 'B', 'C', 'D', 'E', 'F'))}}
    
    @staticmethod
    def n_syms(n): return { f'{i}':i for i in range(n) }

    @staticmethod
    def to_base_10(syms:dict, num:str):
        b = len(syms)
        l = [ syms[n] for n in num[::-1] ]
        return __class__.base2int(l, b)

    @staticmethod
    def from_base_10(syms:dict, num:int, joiner='', ndigs=None):
        base = len(syms)
        #print(f'----{num=} {type(num)}, {base=}, {type(base)}')
        if not ndigs: ndigs = (1 + (0 if num==0 else floor(log(num, base))))  # __class__.ndigs(num, base)
        ss = tuple(syms.keys())
        S = [ ss[i]  for i in __class__.int2base(num, base, ndigs) ]
        return joiner.join(S[::-1])


    @staticmethod
    def int2hex(num:int, joiner=''): return __class__.from_base_10(__class__.SYM_HEX, num, joiner)
  
#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

class Pix(object):
    r""" abstracts a 4-channel (brga) image """

    #=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    CHANNELS = 4
    DTYPE = np.uint8
    CHANNEL_BRGA = (0, 1, 2, 3)
    CHANNEL_RBGA = (1, 2, 0, 3)
    #=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
    def __init__(self, h, w, create=True) -> None:
        self.h, self.w = int(h), int(w)
        if create: self.i = np.zeros((self.h, self.w, __class__.CHANNELS), dtype=__class__.DTYPE) 
    
    @property
    def RBGA(self): return self.i[:, :, __class__.CHANNEL_RBGA]

    def plot_on(self, ax, grid): 
        ax.imshow(self.RBGA)
        if grid:
            xtick, ytick = np.arange(self.w), np.arange(self.h)
            ax.set_xticks(xtick-0.5)
            ax.set_xticklabels(xtick)
            ax.set_yticks(ytick-0.5)
            ax.set_yticklabels(ytick)
            ax.grid(axis='both')

    def clear(self, channel=None): 
        if channel is None:     self.i[:,:,:]       =0 # clear all channels
        else:                   self.i[:,:,channel] =0 # clear specified channel

    def get_color_at(self, row, col, normalize=False):
        b,g,r,a = (self.i[row, col, :]/255 if normalize else self.i[row, col, :])
        return (r, g, b, a)
    
    def set_color_at(self, row:int, col:int, rgba:tuple, normalize=False): 
        if normalize: rgba = [int(x*255) for x in rgba]
        r,g,b,a = rgba
        self.i[row, col, :] = (b, r, g, a) 

    def set_color_in(self, start_row:int, start_col:int, n_rows:int, n_cols:int, rgba:tuple, normalize=False): 
        if normalize: rgba = [int(x*255) for x in rgba]
        r,g,b,a = rgba
        self.i[start_row:start_row+n_rows, start_col:start_col+n_cols, :] = (b, r, g, a) 

    def set_hex_at(self, row:int, col:int, hex:str):
        if hex.startswith('#'): hex = hex[1:]
        hex = hex.upper()
        lenhex = len(hex)
        assert lenhex==6 or lenhex==8, f'expecting 6 or 8 chars but got {lenhex} :: {hex}'
        if lenhex==6: hex = 'FF' + hex # max alpha
        B,G,R,A = tuple(BaseConvertNumpy.int2base(num=BaseConvertNumpy.to_base_10(BaseConvertNumpy.SYM_HEX, hex), base=256, digs=4))
        return self.set_color_at(row,col,(R,G,B,A))

    def set_hex_in(self, start_row:int, start_col:int, n_rows:int, n_cols:int, hex:str):
        if hex.startswith('#'): hex = hex[1:]
        hex = hex.upper()
        lenhex = len(hex)
        assert lenhex==6 or lenhex==8, f'expecting 6 or 8 chars but got {lenhex} :: {hex}'
        if lenhex==6: hex = 'FF' + hex # max alpha
        B,G,R,A = tuple(BaseConvertNumpy.int2base(num=BaseConvertNumpy.to_base_10(BaseConvertNumpy.SYM_HEX, hex), base=256, digs=4))
        return self.set_color_in(start_row,start_col,n_rows,n_cols,(R,G,B,A))



    @staticmethod
    def save(pix, path):  
        return cv2.imwrite(path, pix.i)

    @staticmethod
    def load(path): 
        img =  cv2.imread(path, cv2.IMREAD_UNCHANGED)
        assert img.ndim==3, f'expecting 3-D array but got {img.ndim}-D'
        assert img.shape[-1]== __class__.CHANNELS, f'must be argb image with {__class__.CHANNELS} channels but got {img.shape[-1]} channels'
        h, w, _ = img.shape
        pix = __class__(h, w, False)
        pix.i = img.astype(__class__.DTYPE)
        return pix

    @staticmethod
    def plot(pix, ratio=0.75, grid=True):
        fig, ax = plt.subplots(1,1, figsize=(pix.w*ratio, pix.h*ratio))
        pix.plot_on(ax, grid)
        plt.show()

    def show_color_at(self, row, col, fs=1):
        rgba = self.get_color_at(row,col, normalize=False)
        hexc = self.get_hex_at(row, col)
        color = self.get_color_at(row, col, normalize=True)
        if not isinstance(fs, tuple): fs=(fs,fs)
        plt.figure(figsize=fs)
        plt.yticks([], [])
        plt.title(f'{rgba=}')
        plt.bar([f'{hexc}'], [1], color=color)
        plt.show()
        plt.close()

    @staticmethod
    def region(from_pix, start_row, start_col, n_rows, n_cols):
        r""" creates a new class object from a rectangular region with upper left corner at (x,y) and size (w,h)"""
        pix = __class__(n_rows, n_cols, False)
        pix.i = np.copy(from_pix.i[start_row:start_row+n_rows, start_col:start_col+n_cols,  :])
        return pix

    @staticmethod
    def copy_region(pix_from, start_row, start_col, n_rows, n_cols, pix_to, start_row_to, start_col_to):
        pix_to.i[start_row_to:start_row_to+n_rows, start_col_to:start_col_to+n_cols,  :] = pix_from.i[start_row:start_row+n_rows, start_col:start_col+n_cols,  :]

    @staticmethod
    def copy(pix_from, pix_to): pix_to.i[:, :, :] = pix_from.i[:, :, :]

    def clone(self):
        pix = self.__class__(self.h, self.w, create=False)
        pix.i = np.copy(self.i)
        return pix
        
    @staticmethod
    def graphfromimage(img_path:str, pixel_choice:str='first', dtype=None):
        r""" 
        Covert an image to an array (1-Dimensional)

        :param img_path:        path of input image 
        :param pixel_choice:    choose from ``[ 'first', 'last', 'mid', 'mean' ]``

        :returns: 1-D numpy array containing the data points

        .. note:: 
            * This is used to generate synthetic data in 1-Dimension. 
                The width of the image is the number of points (x-axis),
                while the height of the image is the range of data points, choosen based on their index along y-axis.
        
            * The provided image is opened in grayscale mode.
                All the *black pixels* are considered as data points.
                If there are multiple black points in a column then ``pixel_choice`` argument specifies which pixel to choose.

            * Requires ``opencv-python``

                Input image should be readable using ``cv2.imread``.
                Use ``pip install opencv-python`` to install ``cv2`` package
        """
        img= cv2.imread(img_path, 0)
        imgmax = img.shape[1]-1
        j = img*0
        j[np.where(img==0)]=1
        pixel_choice = pixel_choice.lower()
        pixel_choice_dict = {
            'first':    (lambda ai: ai[0]),
            'last':     (lambda ai: ai[-1]),
            'mid':      (lambda ai: ai[int(len(ai)/2)]),
            'mean':     (lambda ai: np.mean(ai))
        }
        px = pixel_choice_dict[pixel_choice]
        if dtype is None: dtype=np.float_
        return np.array([ imgmax-px(np.where(j[:,i]==1)[0]) for i in range(j.shape[1]) ], dtype=dtype)

#=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=



__doc__=f""" 
-------------------------------------------------------------
topics - Flask-based web app for sharing files 
-------------------------------------------------------------
Usage:
python -m venv .venv
source .venv/bin/activate
python -m pip install Flask Flask-WTF waitress nbconvert
python -m pip install Flask==3.0.2 Flask-WTF==1.2.1 waitress==3.0.0 nbconvert==7.16.2
python topics.py 
-------------------------------------------------------------
Configs:
default `configs.py` file will be created
the dict named `current` will be choosen as the config
it should be defined at the end of the file
a config name `default` will be created 
it is used as a fall-back config
-------------------------------------------------------------
Note:
special string "::::" is used for replacing javascript on `repass` - uid and url should not contain this
special username 'None' is not allowed however words like 'none' will work
rename argument means (0 = not allowed) (1 = only rename) (2 = rename and remoji)
-------------------------------------------------------------


Note:
we use waitress for serving - which uses threads
threads in python are not really concurrent
all we need to do is declare variables as global when accessing from a thread
global variables are shared accross threads and not exclusive to thread
we must declare global variables only when writing to that variable but not on read 
"""






__doc__="""
HTTP API Request-Response model

Client applications will use only the 'POST' method which can be called with 3 types of data

response = requests.post(
    url=server,
    data={key: value}, # A dictionary, list of tuples, bytes or a file object to send to the specified url
    json={key: value}, # A JSON object to send to the specified url
    files={key: value}, # File handles (opened)
    timeout=0.0,
    )

The 'GET' method will be used as 

response = requests.get(
    url=server,
    params={key: value}, # A dictionary of string params in a get url like scheme
    timeout=0.0,
    )


How to design User Module to handle api events

# ----------------------------------------------------------------------
# as global function 
# python -m known.api --user=user.py 
# ----------------------------------------------------------------------
def handle_json(host, data):
    f = f'↪ [Module handle json]\n{host=}\n{data=}\n'
    print(f)
    return f


def handle_file(host, files):
    f = f'↪ [Module handle file]\n{host=}\n{files=}\n'
    print(f)
    return f


# ----------------------------------------------------------------------
# as a global object 
# python -m known.api --user=user.py --object=StaticChan
# ----------------------------------------------------------------------
class StaticChan:

    @staticmethod
    def handle_json(host, data):
        f = f'↪ [{__class__} Module handle json]\n{host=}\n{data=}\n'
        print(f)
        return f
    
    @staticmethod
    def handle_file(host, files):
        f = f'↪ [{__class__} Module handle file]\n{host=}\n{files=}\n'
        print(f)
        return f


# ----------------------------------------------------------------------
# as a global Module (callable - requires initialization) 
# python -m known.api --user=user.py --object=InitChan --callable=1
# ----------------------------------------------------------------------
class InitChan:

    def __init__(self): pass

    def handle_json(self, host, data):
        f = f'↪ [{self.__class__} Module handle json]\n{host=}\n{data=}\n'
        print(f)
        return f
    
    
    def handle_file(self, host, files):
        f = f'↪ [{self.__class__} Module handle file]\n{host=}\n{files=}\n'
        print(f)
        return f
"""

#-----------------------------------------------------------------------------------------
from sys import exit
if __name__!='__main__': exit(f'[!] can not import {__name__}.{__file__}')
#-----------------------------------------------------------------------------------------

# ------------------------------------------------------------------------------------------
# imports ----------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------
import os, argparse, datetime, importlib, importlib.util

#PYDIR = os.path.dirname(__file__) # script directory of __main__.py
try:
    from flask import Flask, request
    from waitress import serve
    from http import HTTPStatus
except: exit(f'[!] The required packages missing:\tFlask>=3.0.2, waitress>=3.0.0\n  ⇒ pip install Flask waitress')


class DefaultUserModule:
    def handle_args(self, host, args):        return f"handle_args({host}, {args})"
    def handle_json(self, host, data):        return f"handle_json({host}, {data})"    
    def handle_file(self, host, files):       return f"handle_file({host}, {files})"


parser = argparse.ArgumentParser()
parser.add_argument('--user', type=str, default='', help="path of main user module file")
parser.add_argument('--object', type=str, default='', help="path of main user module")
parser.add_argument('--callable', type=int, default=0, help="if true, calls the class to create a new instance (with no args)")
parser.add_argument('--handle_json', type=str, default='handle_json', help="path of main user module file")
parser.add_argument('--handle_file', type=str, default='handle_file', help="path of main user module file")
parser.add_argument('--handle_args', type=str, default='handle_args', help="path of main user module file")
parser.add_argument('--host', type=str, default='0.0.0.0', help="host ip-address")
parser.add_argument('--port', type=str, default='8080', help="host port")
parser.add_argument('--max', type=str, default='1GB', help="max_request_body_size")
parser.add_argument('--limit', type=int, default=10, help="connection limit")
parsed = parser.parse_args()

if parsed.user:
    # user-module
    USER_MODULE_FILE_PATH = os.path.abspath(parsed.user)
    print(f'↪ Loading user-module from {USER_MODULE_FILE_PATH}')
    if not os.path.isfile(USER_MODULE_FILE_PATH): exit(f'Invalid user-module file @ {USER_MODULE_FILE_PATH}')
    try: 
        # from https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
        user_module_spec = importlib.util.spec_from_file_location("", USER_MODULE_FILE_PATH)
        user_module = importlib.util.module_from_spec(user_module_spec)
        user_module_spec.loader.exec_module(user_module)
        print(f'↪ Imported user-module from {user_module.__file__}')
    except: exit(f'[!] Could import user-module "{USER_MODULE_FILE_PATH}"')
    if parsed.object:
        try:
            user_module = getattr(user_module, parsed.object)
            if bool(parsed.callable): user_module = user_module()
        except:
            exit(f'Could not load object {parsed.object}')
    USER_HANDLE_JSON = parsed.handle_json
    USER_HANDLE_FILE = parsed.handle_file
    USER_HANDLE_ARGS = parsed.handle_args
else:
    print(f'↪ [!] user-module not defined, using default.')
    USER_HANDLE_JSON = 'handle_json'
    USER_HANDLE_FILE = 'handle_file'
    USER_HANDLE_ARGS = 'handle_args'
    user_module = DefaultUserModule()

if not hasattr(user_module, USER_HANDLE_ARGS): exit(f'[!] ARGS Handler Method not found {user_module}.{USER_HANDLE_ARGS}')
if not hasattr(user_module, USER_HANDLE_JSON): exit(f'[!] JSON Handler Method not found {user_module}.{USER_HANDLE_JSON}')
if not hasattr(user_module, USER_HANDLE_FILE): exit(f'[!] FILE Handler Method not found {user_module}.{USER_HANDLE_FILE}')
#from known.basic import Verbose as vb
# ------------------------------------------------------------------------------------------
# application setting and instance
# ------------------------------------------------------------------------------------------
app = Flask(__name__)
#app.secret_key =         'api_test'
#app.config['key'] =      'value'

# NOTE: "return_object" must be a string, dict, list, tuple with headers or status, Response instance, or WSGI callable

def is_valid_return_object(x): return isinstance(x, (str, dict, list))

@app.route('/', methods =['GET', 'POST'])
def home():
    #vb.show(request)
    if request.method == 'POST':
        #print('FILENAME:', request.files.filename())
        if request.is_json:  
            return_object = getattr(user_module, USER_HANDLE_JSON)(request.host, request.get_json())           
            if is_valid_return_object(return_object): return_code = HTTPStatus.OK
            else: return_object, return_code = f"[!] [Handle-POST-JSON] The Return Type {type(return_object)} is Invalid", HTTPStatus.NOT_FOUND
        else:
            if request.args or request.form: 
                return_object = getattr(user_module, USER_HANDLE_ARGS)(request.host, request.form)   
                if is_valid_return_object(return_object): return_code = HTTPStatus.OK
                else: return_object, return_code = f"[!] [Handle-POST-ARGS] The Return Type {type(return_object)} is Invalid", HTTPStatus.NOT_FOUND
            else:
                if request.files: 
                    return_object = getattr(user_module, USER_HANDLE_FILE)(request.host, request.files) 
                    if is_valid_return_object(return_object): return_code = HTTPStatus.OK
                    else: return_object, return_code = f"[!] [Handle-POST-FILE] The Return Type {type(return_object)} is Invalid", HTTPStatus.NOT_FOUND
                else: return_object, return_code = f"[!] Invalid [POST] Request - Must be json or have files", HTTPStatus.BAD_REQUEST
    elif request.method == 'GET':       
        return_object = f'<pre>[Known.api]@{__file__}\n'
        for k,v in parsed._get_kwargs(): return_object+=f'\n\t{k}\t{v}\n'
        return_object+='</pre>'
        return_code = HTTPStatus.OK
    else:                               return_object, return_code = f"[!] Invalid Request Type {request.method}", HTTPStatus.BAD_REQUEST
    # only use either one of data or json in post request (not both)
    return return_object, return_code

   

#%% @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
#<-------------------DO NOT WRITE ANY CODE AFTER THIS

HRsizes = dict(KB=2**10, MB=2**20, GB=2**30, TB=2**40)
str2bytes = lambda size: int(float(size[:-2])*HRsizes.get(size[-2:].upper(), 0))

start_time = datetime.datetime.now()
print('◉ start server @ [{}]'.format(start_time))
serve(app, # https://docs.pylonsproject.org/projects/waitress/en/stable/runner.html
    host = parsed.host,          
    port = parsed.port,          
    url_scheme = 'http',     
    threads = 1,    
    connection_limit = parsed.limit,
    max_request_body_size = str2bytes(parsed.max),
)
#<-------------------DO NOT WRITE ANY CODE AFTER THIS
end_time = datetime.datetime.now()
print('◉ stop server @ [{}]'.format(end_time))
print('◉ server up-time was [{}]'.format(end_time - start_time))
